# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import numpy as np
import pytest
import biotite.sequence as seq
import biotite.sequence.align as align
from biotite.sequence.seqtypes import ProteinSequence


@pytest.mark.parametrize(
    "seq_type, seq1, seq2, seed, threshold, ref_range1, ref_range2,"
    "direction, score_only, uint8_code",
    [list(itertools.chain(*e)) for e in itertools.product(
        [
            (
                seq.NucleotideSequence,  # seq_type
                "TTTAAAAAAATTTTT",       # seq1
                "CCCCCAAAAAAACCC",       # seq1
                (5, 7),                  # seed
                0,                       # threshold
                (3, 10),                 # ref_range1
                (5, 12),                 # ref_range2
            ),
            (
                seq.NucleotideSequence,
                "TTTAAAAAAATTTTT",
                "CCCCCAAAAAAACCC",
                (3, 5),
                100,
                (3, 10),
                (5, 12),
            ),
            # Mismatch should lead to alignment termination
            (
                seq.ProteinSequence,
                "NLYIQWLKDGGPSSGRPPPS",
                  "YIQWLKDGGPSLGRPPPS",
                (8, 6),
                0,
                (2, 13),
                (0, 11),
            ),
            # Mismatch should not terminate alignment
            (
                seq.ProteinSequence,
                "NLYIQWLKDGGPSSGRPPPS",
                  "YIQWLKDGGPSLGRPPPS",
                (8, 6),
                2,
                (2, 20),
                (0, 18),
            ),
            # Mismatch should terminate alignment
            (
                seq.ProteinSequence,
                "NLYIQWLKDGGPSSGRPPPS",
                  "YIQWLKDWWPSSGRPPPS",
                (6, 4),
                3,
                (2, 9),
                (0, 7),
            ),
        ],

        [["both"], ["upstream"], ["downstream"]],  # direction
        
        [[False], [True]],  # score_only

        [[False], [True]],  # uint8_code
    )]
)
def test_simple_alignments(seq_type, seq1, seq2, seed, threshold,
                           ref_range1, ref_range2,
                           direction, score_only, uint8_code):
    """
    Check if `algin_local_ungapped()` produces correct alignments based on
    simple known examples.
    """
    # Limit start or stop reference alignment range to seed
    # if the alignment does not extend in both directions
    if direction == "upstream":
        ref_range1 = (ref_range1[0], seed[0] + 1)
        ref_range2 = (ref_range2[0], seed[1] + 1)
    elif direction == "downstream":
        ref_range1 = (seed[0], ref_range1[1])
        ref_range2 = (seed[1], ref_range2[1])

    seq1 = seq_type(seq1)
    seq2 = seq_type(seq2)
    
    if seq_type == seq.NucleotideSequence:
        matrix = align.SubstitutionMatrix.std_nucleotide_matrix()
    else:
        matrix = align.SubstitutionMatrix.std_protein_matrix()
    
    if not uint8_code:
        seq1, seq2, matrix = _convert_to_uint16_code(seq1, seq2, matrix)

    
    ref_alignment = align.Alignment(
        [seq1, seq2],
        np.stack([
            np.arange(*ref_range1),
            np.arange(*ref_range2)
        ], axis=-1)
    )
    ref_score = align.score(ref_alignment, matrix)
    ref_alignment.score = ref_score

    test_result = align.align_local_ungapped(
        seq1, seq2, matrix, seed, threshold, direction, score_only)
    
    if score_only:
        assert test_result == ref_score
    else:
        assert test_result == ref_alignment


@pytest.mark.parametrize(
    "seed, uint8_code", itertools.product(
        range(100),
        [False, True]
    )
)
def test_random_alignment(seed, uint8_code):
    """
    Create two randomized sequences and place a conserved region into
    each sequence, where both conserved regions are similar to each
    other.
    The conserved regions only contain point mutations and no indels.
    Expect that the alignment score found by `align_local_ungapped()` is
    equal to the alignment score found by `align_optimal()`.
    """
    MIN_SIZE = 200
    MAX_SIZE = 1000
    MIN_CONSERVED_SIZE = 20
    MAX_CONSERVED_SIZE = 100
    CONSERVED_ENDS = 5
    MUTATION_PROB = 0.1
    THRESHOLD = 100
    
    np.random.seed(seed)

    # Create conserved regions
    conserved1 = ProteinSequence()
    conserved_len = np.random.randint(MIN_CONSERVED_SIZE, MAX_CONSERVED_SIZE+1)
    conserved1.code = np.random.randint(
        # Do not include stop symbol for aesthetic reasons -> -1
        len(conserved1.alphabet)-1,
        size=conserved_len
    )
    conserved2 = ProteinSequence()
    # The second conserved regions is equal to the first one,
    # except a few point mutations
    conserved2.code = conserved1.code.copy()
    mutation_mask = np.random.choice(
        [False, True],
        size=conserved_len,
        p = [1 - MUTATION_PROB, MUTATION_PROB]
    )
    conserved2.code[mutation_mask] = np.random.randint(
        len(conserved2.alphabet)-1,
        size=np.count_nonzero(mutation_mask)
    )
    # Flank the conserved regions with equal termini to ensure
    # that the alignment extends from start to end of the region
    conserved2.code[:CONSERVED_ENDS] = conserved1.code[:CONSERVED_ENDS]
    conserved2.code[-CONSERVED_ENDS:] = conserved1.code[-CONSERVED_ENDS:]

    # Create randomized sequences
    seq1 = ProteinSequence()
    seq2 = ProteinSequence()
    offset = []
    for sequence, conserved in zip(
        (seq1, seq2), (conserved1, conserved2)
    ):
        sequence.code = np.random.randint(
            len(sequence.alphabet)-1,
            size=np.random.randint(MIN_SIZE, MAX_SIZE+1)
        )
        # Place conserved region randomly within the sequence
        conserved_pos = np.random.randint(0, len(sequence) - len(conserved))
        sequence.code[conserved_pos : conserved_pos + len(conserved)] \
            = conserved.code
        offset.append(conserved_pos)
    # The seed is placed somewhere in the conserved region
    seed = np.array(offset) + np.random.randint(len(conserved))


    matrix = align.SubstitutionMatrix.std_protein_matrix()
    if not uint8_code:
        seq1, seq2, matrix = _convert_to_uint16_code(seq1, seq2, matrix)
    
    ref_score = align.align_optimal(
        seq1, seq2, matrix, local=True, max_number=1,
        # High gap penalty to prevent introduction of gaps, 
        # since 'align_local_ungapped()' is also no able to place gaps
        gap_penalty=-1000
    )[0].score

    test_alignment = align.align_local_ungapped(
        seq1, seq2, matrix, seed, THRESHOLD
    )

    assert test_alignment.score == ref_score
    # Test if the score is also correctly calculated
    assert align.score(test_alignment, matrix) == ref_score


def _convert_to_uint16_code(seq1, seq2, matrix):
        """
        Adjust sequences, so that they use 'uint16' as dtype for the
        code.
        This is a necessary test, since 'uint8' uses a separate
        implementation.
        """
        new_alph = seq.Alphabet(np.arange(500))
        code = seq1.code
        seq1 = seq.GeneralSequence(new_alph)
        seq1.code = code
        code = seq2.code
        seq2 = seq.GeneralSequence(new_alph)
        seq2.code = code
        # Adjust the substitution matrix as well,
        # so that it is compatible with the new alphabet
        score_matrix = np.zeros((len(new_alph), len(new_alph)), dtype=np.int32)
        orig_len = len(matrix.score_matrix())
        score_matrix[:orig_len, :orig_len] = matrix.score_matrix()
        matrix = align.SubstitutionMatrix(new_alph, new_alph, score_matrix)
        return seq1, seq2, matrix


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing sequencing data
using the popular FASTQ format.

This package contains the :class:`FastqFile`, which provides a
dictionary like interface to FASTQ files, with the sequence identifer
strings being the keys and the sequences and quality scores being the
values.
"""

__name__ = "biotite.sequence.io.fastq"
__author__ = "Patrick Kunzmann"

from .file import *
from .convert import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.viennarna"
__author__ = "Tom David Müller"
__all__ = ["RNAfoldApp"]

from tempfile import NamedTemporaryFile
from ..localapp import LocalApp, cleanup_tempfile
from ..application import AppState, requires_state
from ...sequence.io.fasta import FastaFile, set_sequence
from ...sequence import NucleotideSequence
from ...structure.dotbracket import base_pairs_from_dot_bracket

class RNAfoldApp(LocalApp):
    """
    Compute the minimum free energy secondary structure of a ribonucleic
    acid sequence using *ViennaRNA's* *RNAfold* software.

    Internally this creates a :class:`Popen` instance, which handles
    the execution.

    Parameters
    ----------
    sequence : NucleotideSequence
        The RNA sequence.
    temperature : int, optional
        The temperature (°C) to be assumed for the energy parameters.
    bin_path : str, optional
        Path of the *RNAfold* binary.

    Examples
    --------

    >>> sequence = NucleotideSequence("CGACGTAGATGCTAGCTGACTCGATGC")
    >>> app = RNAfoldApp(sequence)
    >>> app.start()
    >>> app.join()
    >>> print(app.get_mfe())
    -1.3
    >>> print(app.get_dot_bracket())
    (((.((((.......)).)))))....
    """

    def __init__(self, sequence, temperature=37, bin_path="RNAfold"):
        super().__init__(bin_path)
        self._sequence = sequence
        self._in_file  = NamedTemporaryFile("w", suffix=".fa",  delete=False)
        self._temperature = str(temperature)

    def run(self):
        in_file = FastaFile()
        set_sequence(in_file, self._sequence)
        in_file.write(self._in_file)
        self._in_file.flush()
        self.set_arguments(
            [self._in_file.name, "--noPS", "-T", self._temperature]
        )
        super().run()

    def evaluate(self):
        super().evaluate()
        lines = self.get_stdout().split("\n")
        content = lines[2]
        dotbracket, mfe = content.split(" ", maxsplit=1)
        mfe = float(mfe[1:-1])

        self._mfe = mfe
        self._dotbracket = dotbracket

    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._in_file)

    @requires_state(AppState.CREATED)
    def set_temperature(self, temperature):
        """
        Adjust the energy parameters according to a temperature in
        degrees Celsius.

        Parameters
        ----------
        temperature : int
            The temperature.
        """
        self._temperature = str(temperature)

    @requires_state(AppState.JOINED)
    def get_mfe(self):
        """
        Get the minimum free energy of the input sequence.

        Returns
        -------
        mfe : float
            The minimum free energy.

        Examples
        --------

        >>> sequence = NucleotideSequence("CGACGTAGATGCTAGCTGACTCGATGC")
        >>> app = RNAfoldApp(sequence)
        >>> app.start()
        >>> app.join()
        >>> print(app.get_mfe())
        -1.3
        """
        return self._mfe

    @requires_state(AppState.JOINED)
    def get_dot_bracket(self):
        """
        Get the minimum free energy secondary structure of the input
        sequence in dot bracket notation.

        Returns
        -------
        dotbracket : str
            The secondary structure in dot bracket notation.

        Examples
        --------

        >>> sequence = NucleotideSequence("CGACGTAGATGCTAGCTGACTCGATGC")
        >>> app = RNAfoldApp(sequence)
        >>> app.start()
        >>> app.join()
        >>> print(app.get_dot_bracket())
        (((.((((.......)).)))))....
        """
        return self._dotbracket

    @requires_state(AppState.JOINED)
    def get_base_pairs(self):
        """
        Get the base pairs from the minimum free energy secondary
        structure of the input sequence.

        Returns
        -------
        base_pairs : ndarray, shape=(n,2)
            Each row corresponds to the positions of the bases in the
            sequence.

        Examples
        --------

        >>> sequence = NucleotideSequence("CGACGTAGATGCTAGCTGACTCGATGC")
        >>> app = RNAfoldApp(sequence)
        >>> app.start()
        >>> app.join()
        >>> print(app.get_base_pairs())
            [[ 0 22]
             [ 1 21]
             [ 2 20]
             [ 4 19]
             [ 5 18]
             [ 6 16]
             [ 7 15]]

        For reference, the corresponding dot bracket notation can be
        displayed as below.

        >>> print(app.get_dot_bracket())
        (((.((((.......)).)))))....
        """
        return base_pairs_from_dot_bracket(self._dotbracket)

    @staticmethod
    def compute_secondary_structure(sequence, bin_path="RNAfold"):
        """
        Compute the minimum free energy secondary structure of a 
        ribonucleic acid sequence using *ViennaRNA's* *RNAfold* software.

        This is a convenience function, that wraps the
        :class:`RNAfoldApp` execution.

        Parameters
        ----------
        sequence : NucleotideSequence
            The RNA sequence.
        bin_path : str, optional
            Path of the *RNAfold* binary.

        Returns
        -------
        dotbracket : str
            The secondary structure in dot bracket notation.
        mfe : float
            The minimum free energy.
        """
        app = RNAfoldApp(sequence, bin_path=bin_path)
        app.start()
        app.join()
        return app.get_dot_bracket(), app.get_mfe()



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.application.dssp"
__author__ = "Patrick Kunzmann"
__all__ = ["DsspApp"]

from tempfile import NamedTemporaryFile
from ..localapp import LocalApp, cleanup_tempfile
from ..application import AppState, requires_state
from ...structure.io.pdbx.file import PDBxFile
from ...structure.io.pdbx.convert import set_structure
import numpy as np


class DsspApp(LocalApp):
    r"""
    Annotate the secondary structure of a protein structure using the
    *DSSP* software.
    
    Internally this creates a :class:`Popen` instance, which handles
    the execution.
    
    DSSP differentiates between 8 different types of secondary
    structure elements:
    
       - C: loop, coil or irregular
       - H: :math:`{\alpha}`-helix
       - B: :math:`{\beta}`-bridge
       - E: extended strand, participation in :math:`{\beta}`-ladder
       - G: 3 :sub:`10`-helix
       - I: :math:`{\pi}`-helix
       - T: hydrogen bonded turn
       - S: bend 
    
    Parameters
    ----------
    atom_array : AtomArray
        The atom array to be annotated.
    bin_path : str, optional
        Path of the *DDSP* binary.
    
    Examples
    --------

    >>> app = DsspApp(atom_array)
    >>> app.start()
    >>> app.join()
    >>> print(app.get_sse())
    ['C' 'H' 'H' 'H' 'H' 'H' 'H' 'H' 'T' 'T' 'G' 'G' 'G' 'G' 'T' 'C' 'C' 'C'
     'C' 'C']
    """
    
    def __init__(self, atom_array, bin_path="mkdssp"):
        super().__init__(bin_path)

        # mkdssp requires also the
        # 'occupancy', 'b_factor' and 'charge' fields
        # -> Add these annotations to a copy of the input structure
        self._array = atom_array.copy()
        categories = self._array.get_annotation_categories()
        if "charge" not in categories:
            self._array.set_annotation(
                "charge", np.zeros(self._array.array_length(), dtype=int)
            )
        if "b_factor" not in categories:
            self._array.set_annotation(
                "b_factor", np.zeros(self._array.array_length(), dtype=float)
            )
        if "occupancy" not in categories:
            self._array.set_annotation(
                "occupancy", np.ones(self._array.array_length(), dtype=float)
            )

        self._in_file  = NamedTemporaryFile("w", suffix=".cif",  delete=False)
        self._out_file = NamedTemporaryFile("r", suffix=".dssp", delete=False)

    def run(self):
        in_file = PDBxFile()
        set_structure(in_file, self._array, data_block="DSSP_INPUT")
        in_file.write(self._in_file)
        self._in_file.flush()
        self.set_arguments(
            ["-i", self._in_file.name, "-o", self._out_file.name]
        )
        super().run()
    
    def evaluate(self):
        super().evaluate()
        lines = self._out_file.read().split("\n")
        # Index where SSE records start
        sse_start = None
        for i, line in enumerate(lines):
            if line.startswith("  #  RESIDUE AA STRUCTURE"):
                sse_start = i+1
        if sse_start is None:
            raise ValueError("DSSP file does not contain SSE records")
        lines = [line for line in lines[sse_start:] if len(line) != 0]
        self._sse = np.zeros(len(lines), dtype="U1")
        # Parse file for SSE letters
        for i, line in enumerate(lines):
            self._sse[i] = line[16]
        # Remove "!" for missing residues
        self._sse = self._sse[self._sse != "!"]
        self._sse[self._sse == " "] = "C"
    
    def clean_up(self):
        super().clean_up()
        cleanup_tempfile(self._in_file)
        cleanup_tempfile(self._out_file)
    
    @requires_state(AppState.JOINED)
    def get_sse(self):
        """
        Get the resulting secondary structure assignment.
        
        Returns
        -------
        sse : ndarray, dtype="U1"
            An array containing DSSP secondary structure symbols
            corresponding to the residues in the input atom array.
        """
        return self._sse
    
    @staticmethod
    def annotate_sse(atom_array, bin_path="mkdssp"):
        """
        Perform a secondary structure assignment to an atom array.
        
        This is a convenience function, that wraps the :class:`DsspApp`
        execution.
        
        Parameters
        ----------
        atom_array : AtomArray
            The atom array to be annotated.
        bin_path : str, optional
            Path of the DDSP binary.
        
        Returns
        -------
        sse : ndarray, dtype="U1"
            An array containing DSSP secondary structure symbols
            corresponding to the residues in the input atom array.
        """
        app = DsspApp(atom_array, bin_path)
        app.start()
        app.join()
        return app.get_sse()



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.xtc"
__author__ = "Patrick Kunzmann"
__all__ = ["XTCFile"]

import numpy as np
from ..trajfile import TrajectoryFile


class XTCFile(TrajectoryFile):
    """
    This file class represents a XTC trajectory file.
    """
    
    @classmethod
    def traj_type(cls):
        import mdtraj.formats as traj
        return traj.XTCTrajectoryFile

    @classmethod
    def process_read_values(cls, read_values):
        # nm to Angstrom
        coord = read_values[0] * 10
        box = read_values[3]
        if box is not None:
            box *= 10
        time = read_values[1]
        return coord, box, time
    
    @classmethod
    def prepare_write_values(cls, coord, box, time):
        # Angstrom to nm
        xyz = np.divide(coord, 10, dtype=np.float32) \
              if coord is not None else None
        time = time.astype(np.float32, copy=False) \
               if time is not None else None
        box = np.divide(box, 10, dtype=np.float32) \
              if box is not None else None
        return {
            "xyz"  : xyz,
            "box"  : box,
            "time" : time,
        }



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.io.fastq"
__author__ = "Patrick Kunzmann"

import warnings
from numbers import Integral
from collections import OrderedDict
from collections.abc import MutableMapping
import numpy as np
from ....file import TextFile, InvalidFileError, wrap_string
from ...seqtypes import NucleotideSequence

__all__ = ["FastqFile"]


_OFFSETS = {
    "Sanger"       : 33,
    "Solexa"       : 64,
    "Illumina-1.3" : 64,
    "Illumina-1.5" : 64,
    "Illumina-1.8" : 33,
}


class FastqFile(TextFile, MutableMapping):
    """
    This class represents a file in FASTQ format.

    A FASTQ file stores one or multiple sequences (base calls) along
    with sequencing quality scores.
    Each sequence is associated with an identifer string,
    beginning with an ``@``.

    The quality scores are encoded as ASCII characters,
    with each actual score being the ASCII code subtracted by an
    `offset` value.
    The offset is format dependent.
    As the offset is not reliably deducible from the file contets, it
    must be provided explicitly, either as number or format
    (e.g. ``'Illumina-1.8'``).

    Similar to the :class:`FastaFile` class, this class implements the
    :class:`MutableMapping` interface:
    An identifier string (without the leading ``@``) is used as index
    to get and set the corresponding sequence and quality.
    ``del`` removes an entry in the file.
    
    Parameters
    ----------
    offset : int or {'Sanger', 'Solexa', 'Illumina-1.3', 'Illumina-1.5', 'Illumina-1.8'}
        This value is added to the quality score to obtain the
        ASCII code.
        Can either be directly the value, or a string that indicates
        the score format.
    chars_per_line : int, optional
        The number characters in a line containing sequence data
        after which a line break is inserted.
        Only relevant, when adding sequences to a file.
        By default each sequence (and score string)
        is put into one line.
    
    Examples
    --------
    
    >>> import os.path
    >>> file = FastqFile(offset="Sanger")
    >>> file["seq1"] = str(NucleotideSequence("ATACT")), [0,3,10,7,12]
    >>> file["seq2"] = str(NucleotideSequence("TTGTAGG")), [15,13,24,21,28,38,35]
    >>> print(file)
    @seq1
    ATACT
    +
    !$+(-
    @seq2
    TTGTAGG
    +
    0.96=GD
    >>> sequence, scores = file["seq1"]
    >>> print(sequence)
    ATACT
    >>> print(scores)
    [ 0  3 10  7 12]
    >>> del file["seq1"]
    >>> print(file)
    @seq2
    TTGTAGG
    +
    0.96=GD
    >>> file.write(os.path.join(path_to_directory, "test.fastq"))
    """
    
    def __init__(self, offset, chars_per_line=None):
        super().__init__()
        self._chars_per_line = chars_per_line
        self._entries = OrderedDict()
        self._offset = _convert_offset(offset)
    
    @classmethod
    def read(cls, file, offset, chars_per_line=None):
        """
        Read a FASTQ file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        offset : int or {'Sanger', 'Solexa', 'Illumina-1.3', 'Illumina-1.5', 'Illumina-1.8'}
            This value is added to the quality score to obtain the
            ASCII code.
            Can either be directly the value, or a string that indicates
            the score format.
        chars_per_line : int, optional
            The number characters in a line containing sequence data
            after which a line break is inserted.
            Only relevant, when adding sequences to a file.
            By default each sequence (and score string)
            is put into one line.
        
        Returns
        -------
        file_object : FastqFile
            The parsed file.
        """
        file = super().read(file, offset, chars_per_line)
        # Remove leading and trailing whitespace in all lines
        file.lines = [line.strip() for line in file.lines]
        # Filter out empty lines
        file.lines = [line for line in file.lines if len(line) != 0]
        if len(file.lines) == 0:
            raise InvalidFileError("File is empty")
        file._find_entries()
        return file
    
    def get_sequence(self, identifier):
        """
        Get the sequence for the specified identifier.

        DEPRECATED: Use :meth:`get_seq_string()` or
        :func:`get_sequence()` instead.

        Parameters
        ----------
        identifier : str
            The identifier of the sequence.
        
        Returns
        -------
        sequence : NucleotideSequence
            The sequence corresponding to the identifier.
        """
        warnings.warn(
            "'get_sequence()' is deprecated, use the 'get_seq_string()'"
            "method or 'fasta.get_sequence()' function instead",
            DeprecationWarning
        )
        return NucleotideSequence(self.get_seq_string(identifier))
    
    def get_seq_string(self, identifier):
        """
        Get the string representing the sequence for the specified
        identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the sequence.
        
        Returns
        -------
        sequence : str
            The sequence corresponding to the identifier.
        """
        if not isinstance(identifier, str):
            raise IndexError(
                "'FastqFile' only supports identifier strings as keys"
            )
        seq_start, seq_stop, score_start, score_stop \
            = self._entries[identifier]
        # Concatenate sequence string from the sequence lines
        seq_str = "".join(self.lines[seq_start : seq_stop])
        return seq_str
    
    def get_quality(self, identifier):
        """
        Get the quality scores for the specified identifier.

        Parameters
        ----------
        identifier : str
            The identifier of the quality scores.
        
        Returns
        -------
        scores : ndarray, dtype=int
            The quality scores corresponding to the identifier.
        """
        if not isinstance(identifier, str):
            raise IndexError(
                "'FastqFile' only supports identifier strings as keys"
            )
        seq_start, seq_stop, score_start, score_stop \
            = self._entries[identifier]
        # Concatenate sequence string from the score lines
        return _score_str_to_scores(
            "".join(self.lines[score_start : score_stop]),
            self._offset
        )
        
    def __setitem__(self, identifier, item):
        sequence, scores = item
        if len(sequence) != len(scores):
            raise ValueError(
                f"Sequence has length {len(sequence)}, "
                f"but score length is {len(scores)}"
            )
        if not isinstance(identifier, str):
            raise IndexError(
                "'FastqFile' only supports strings as identifier"
            )
        # Delete lines of entry corresponding to the identifier,
        # if already existing
        if identifier in self:
            del self[identifier]
        
        # Create new lines
        # Start with identifier line
        new_lines = ["@" + identifier.replace("\n","").strip()]
        # Append new lines with sequence string (with line breaks)
        seq_start_i = len(new_lines)
        if self._chars_per_line is None:
            new_lines.append(str(sequence))
        else:
            new_lines += wrap_string(sequence, width=self._chars_per_line)
        seq_stop_i =len(new_lines)
        # Append sequence-score separator
        new_lines += ["+"]
        # Append scores
        score_chars = _scores_to_score_str(scores, self._offset)
        score_start_i = len(new_lines)
        if self._chars_per_line is None:
            new_lines.append(score_chars)
        else:
            new_lines += wrap_string(score_chars, width=self._chars_per_line)
        score_stop_i = len(new_lines)

        if identifier in self:
            # Delete lines of entry corresponding to the header,
            # if existing
            del self[identifier]
            self.lines += new_lines
            self._find_entries()
        else:
            # Simply append lines
            # Add entry in a more efficient way than '_find_entries()'
            # for this simple case
            self._entries[identifier] = (
                len(self.lines) + seq_start_i,
                len(self.lines) + seq_stop_i,
                len(self.lines) + score_start_i,
                len(self.lines) + score_stop_i
            )
            self.lines += new_lines
    
    def __getitem__(self, identifier):
        return self.get_seq_string(identifier), self.get_quality(identifier)
    
    def __delitem__(self, identifier):
        seq_start, seq_stop, score_start, score_stop \
            = self._entries[identifier]
        del self.lines[seq_start-1 : score_stop]
        del self._entries[identifier]
        self._find_entries()
    
    def __len__(self):
        return len(self._entries)
    
    def __iter__(self):
        return self._entries.__iter__()
    
    def __contains__(self, identifer):
        return identifer in self._entries
    
    def _find_entries(self):
        self._entries = OrderedDict()
        in_sequence = False
        # Record if the parser is currently in a quality score section,
        # as the '@' character at the start of a line may also be a
        # score instead of the start of an identifier
        in_scores = False
        seq_len = 0
        score_len = 0
        seq_start_i = None
        seq_stop_i = None
        score_start_i = None
        score_stop_i = None
        identifier = None
        for i, line in enumerate(self.lines):
            if not in_scores and not in_sequence and line[0] == "@":
                # Identifier line
                identifier = line[1:]
                seq_start_i = i+1
                # Next line is sequence
                in_sequence = True
                # Reset
                seq_len = 0
                score_len = 0
            elif in_sequence:
                if line[0] == "+":
                    # End of sequence start of scores
                    in_sequence = False
                    in_scores = True
                    seq_stop_i = i
                    score_start_i = i+1
                else:
                    # Still in sequence
                    seq_len += len(line)
            elif in_scores:
                score_len += len(line)
                if score_len < seq_len:
                    # Scores have not ended yet
                    pass
                elif score_len == seq_len:
                    # End of scores
                    # -> End of entry
                    score_stop_i = i + 1
                    in_scores = False
                    # Record this entry
                    self._entries[identifier] = (
                        seq_start_i, seq_stop_i, score_start_i, score_stop_i
                    )
                else: # score_len > seq_len
                    raise InvalidFileError(
                        f"The amount of scores is not equal to the sequence "
                        f"length for the sequence in line {seq_start_i+1} "
                    )
            else:
                raise InvalidFileError(f"Line {i+1} in FASTQ file is invalid")
        # At the end of the file, the last sequence or score block
        # must have properly ended
        if in_sequence or in_scores:
            raise InvalidFileError("The last entry in the file is incomplete")
    

    @staticmethod
    def read_iter(file, offset):
        """
        Create an iterator over each sequence (and corresponding scores)
        of the given FASTQ file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        offset : int or {'Sanger', 'Solexa', 'Illumina-1.3', 'Illumina-1.5', 'Illumina-1.8'}
            This value that is added to the quality score to obtain the
            ASCII code.
            Can either be directly the value, or a string that indicates
            the score format.
        
        Yields
        ------
        identifier : str
            The identifier of the current sequence.
        sequence : tuple(str, ndarray)
            The current sequence as string and its corresponding quality
            scores as :class:`ndarray`.
        
        Notes
        -----
        This approach gives the same results as
        `FastqFile.read(file, offset).items()`, but is slightly faster
        and much more memory efficient.
        """
        offset = _convert_offset(offset)
        
        identifier = None
        seq_str_list = []
        score_str_list = []
        in_sequence = False
        in_scores = False
        seq_len = 0
        score_len = 0

        for line in TextFile.read_iter(file):
            line = line.strip()
            # Ignore empty lines
            if len(line) == 0:
                continue
            
            if not in_scores and not in_sequence and line[0] == "@":
                # Track new entry
                identifier = line[1:]
                in_sequence = True
                # Reset
                seq_len = 0
                score_len = 0
                seq_str_list = []
                score_str_list = []
            
            elif in_sequence:
                if line[0] == "+":
                    # End of sequence start of scores
                    in_sequence = False
                    in_scores = True
                else:
                    # Still in sequence
                    seq_len += len(line)
                    seq_str_list.append(line)
            
            elif in_scores:
                score_len += len(line)
                score_str_list.append(line)
                if score_len < seq_len:
                    pass
                elif score_len == seq_len:
                    # End of scores
                    # -> End of entry
                    in_scores = False
                    # yield this entry
                    scores = _score_str_to_scores(
                        "".join(score_str_list),
                        offset
                    )
                    yield identifier, ("".join(seq_str_list), scores)
                else: # score_len > seq_len
                    raise InvalidFileError(
                        f"The amount of scores is not equal to the sequence "
                        f"length"
                    )
            
            else:
                raise InvalidFileError(f"FASTQ file is invalid")
    

    @staticmethod
    def write_iter(file, items, offset, chars_per_line=None):
        """
        Iterate over the given `items` and write each item into
        the specified `file`.

        In contrast to :meth:`write()`, the lines of text are not stored
        in an intermediate :class:`TextFile`, but are directly written
        to the file.
        Hence, this static method may save a large amount of memory if
        a large file should be written, especially if the `items`
        are provided as generator.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be written to.
            Alternatively a file path can be supplied.
        items : generator or array-like of tuple(str, tuple(str, ndarray))
            The entries to be written into the file.
            Each entry consists of an identifier string and a tuple
            containing a sequence (as string) and a score array.
        offset : int or {'Sanger', 'Solexa', 'Illumina-1.3', 'Illumina-1.5', 'Illumina-1.8'}
            This value is added to the quality score to obtain the
            ASCII code.
            Can either be directly the value, or a string that indicates
            the score format.
        chars_per_line : int, optional
            The number characters in a line containing sequence data
            after which a line break is inserted.
            Only relevant, when adding sequences to a file.
            By default each sequence (and score string)
            is put into one line.

        Notes
        -----
        This method does not test, whether the given identifiers are
        unambiguous.
        """
        offset = _convert_offset(offset)

        def line_generator():
            for item in items:
                identifier, (sequence, scores) = item
                if len(sequence) != len(scores):
                    raise ValueError(
                        f"Sequence has length {len(sequence)}, "
                        f"but score length is {len(scores)}"
                    )
                if not isinstance(identifier, str):
                    raise IndexError(
                        "'FastqFile' only supports strings as identifier"
                    )
                
                # Yield identifier line
                yield "@" + identifier.replace("\n","").strip()

                # Yield sequence line(s)
                if chars_per_line is None:
                    yield str(sequence)
                else:
                    for line in wrap_string(sequence, width=chars_per_line):
                        yield line
                
                # Yield separator
                yield "+"
                
                # Yield scores
                score_chars = _scores_to_score_str(scores, offset)
                if chars_per_line is None:
                    yield score_chars
                else:
                    for line in wrap_string(score_chars, width=chars_per_line):
                        yield line
    
        TextFile.write_iter(file, line_generator())


def _score_str_to_scores(score_str, offset):
    """
    Convert an ASCII string into actual score values.
    """
    scores = np.frombuffer(
        bytearray(
            score_str, encoding="ascii"
        ),
        dtype=np.int8
    )
    scores -= offset
    return scores

def _scores_to_score_str(scores, offset):
    """
    Convert score values into an ASCII string.
    """
    scores = np.asarray(scores) + offset
    return scores.astype(np.int8, copy=False).tobytes().decode("ascii")

def _convert_offset(offset_val_or_string):
    """
    If the given offset is a string return the corresponding numerical
    value.
    """
    if isinstance(offset_val_or_string, Integral):
        return offset_val_or_string
    elif isinstance(offset_val_or_string, str):
       return _OFFSETS[offset_val_or_string]
    else:
        raise TypeError(
            f"The offset must be either an integer or a string "
            f"indicating the format, not {type(offset_val_or_string).__name__}"
        )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.mol"
__author__ = "Patrick Kunzmann"
__all__ = ["MOLFile"]

import datetime
from warnings import warn
import numpy as np
from ...atoms import AtomArray
from ....file import TextFile, InvalidFileError
from ...error import BadStructureError
from ..ctab import read_structure_from_ctab, write_structure_to_ctab


# Number of header lines
N_HEADER = 3
DATE_FORMAT = "%d%m%y%H%M"


class MOLFile(TextFile):
    """
    This class represents a file in MOL format, that is used to store
    structure information for small molecules. :footcite:`Dalby1992`

    Since its use is intended for single small molecules, it stores
    less atom annotation information than the macromolecular structure
    formats:
    Only the atom positions, charges, elements and bonds can be read
    from the file, chain and and residue information is missing.

    This class can also be used to parse the first structure from an SDF
    file, as the SDF format extends the MOL format.
    
    References
    ----------
    
    .. footbibliography::
    
    Examples
    --------

    >>> from os.path import join
    >>> mol_file = MOLFile.read(join(path_to_structures, "molecules", "TYR.sdf"))
    >>> atom_array = mol_file.get_structure()
    >>> print(atom_array)
                0             N         1.320    0.952    1.428
                0             C        -0.018    0.429    1.734
                0             C        -0.103    0.094    3.201
                0             O         0.886   -0.254    3.799
                0             C        -0.274   -0.831    0.907
                0             C        -0.189   -0.496   -0.559
                0             C         1.022   -0.589   -1.219
                0             C        -1.324   -0.102   -1.244
                0             C         1.103   -0.282   -2.563
                0             C        -1.247    0.210   -2.587
                0             C        -0.032    0.118   -3.252
                0             O         0.044    0.420   -4.574
                0             O        -1.279    0.184    3.842
                0             H         1.977    0.225    1.669
                0             H         1.365    1.063    0.426
                0             H        -0.767    1.183    1.489
                0             H         0.473   -1.585    1.152
                0             H        -1.268   -1.219    1.134
                0             H         1.905   -0.902   -0.683
                0             H        -2.269   -0.031   -0.727
                0             H         2.049   -0.354   -3.078
                0             H        -2.132    0.523   -3.121
                0             H        -0.123   -0.399   -5.059
                0             H        -1.333   -0.030    4.784
    """
    
    def __init__(self):
        super().__init__()
        # empty header lines
        self.lines = [""] * N_HEADER
    
    def get_header(self):
        """
        Get the header from the MOL file.
        
        Returns
        -------
        mol_name : str
            The name of the molecule.
        initials : str
            The author's initials.
        program : str
            The program name.
        time : datetime
            The time of file creation.
        dimensions : str
            Dimensional codes.
        scaling_factors : str
            Scaling factors.
        energy : str
            Energy from modeling program.
        registry_number : str
            MDL registry number.
        comments : str
            Additional comments.
        """
        mol_name        = self.lines[0].strip()
        initials        = self.lines[1][0:2].strip()
        program         = self.lines[1][2:10].strip()
        time            = datetime.datetime.strptime(self.lines[1][10:20],
                                                     DATE_FORMAT)
        dimensions      = self.lines[1][20:22].strip()
        scaling_factors = self.lines[1][22:34].strip()
        energy          = self.lines[1][34:46].strip()
        registry_number = self.lines[1][46:52].strip()
        comments        = self.lines[2].strip()
        return mol_name, initials, program, time, dimensions, \
               scaling_factors, energy, registry_number, comments


    def set_header(self, mol_name, initials="", program="", time=None,
                   dimensions="", scaling_factors="", energy="",
                   registry_number="", comments=""):
        """
        Set the header for the MOL file.
        
        Parameters
        ----------
        mol_name : str
            The name of the molecule.
        initials : str, optional
            The author's initials. Maximum length is 2.
        program : str, optional
            The program name. Maximum length is 8.
        time : datetime or date, optional
            The time of file creation.
        dimensions : str, optional
            Dimensional codes. Maximum length is 2.
        scaling_factors : str, optional
            Scaling factors. Maximum length is 12.
        energy : str, optional
            Energy from modeling program. Maximum length is 12.
        registry_number : str, optional
            MDL registry number. Maximum length is 6.
        comments : str, optional
            Additional comments.
        """
        if time is None:
            time = datetime.datetime.now()
        time_str = time.strftime(DATE_FORMAT)

        self.lines[0] = str(mol_name)
        self.lines[1] = (
            f"{initials:>2}"
            f"{program:>8}"
            f"{time_str:>10}"
            f"{dimensions:>2}"
            f"{scaling_factors:>12}"
            f"{energy:>12}"
            f"{registry_number:>6}"
        )
        self.lines[2] = str(comments)


    def get_structure(self):
        """
        Get an :class:`AtomArray` from the MOL file.
        
        Returns
        -------
        array : AtomArray
            This :class:`AtomArray` contains the optional ``charge``
            annotation and has an associated :class:`BondList`.
            All other annotation categories, except ``element`` are
            empty.
        """
        ctab_lines = _get_ctab_lines(self.lines)
        if len(ctab_lines) == 0:
            raise InvalidFileError("File does not contain structure data") 
        return read_structure_from_ctab(ctab_lines)
    

    def set_structure(self, atoms):
        """
        Set the :class:`AtomArray` for the file.
        
        Parameters
        ----------
        array : AtomArray
            The array to be saved into this file.
            Must have an associated :class:`BondList`.
        """
        self.lines = self.lines[:N_HEADER] + write_structure_to_ctab(atoms)


def _get_ctab_lines(lines):
    for i, line in enumerate(lines):
        if line.startswith("M  END"):
            return lines[N_HEADER:i+1]
    return lines[N_HEADER:]


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
The MOL format is used to depict atom positions and bonds for small
molecules.
This subpackage is used for reading and writing an :class:`AtomArray`
in this format.
Additionally, reading data from the SDF format, which is a wrapper
around MOL, is also supported.
"""

__name__ = "biotite.structure.io.mol"
__author__ = "Patrick Kunzmann"

from .file import *
from .convert import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import glob
from os.path import join, splitext
import numpy as np
import pytest
from pytest import approx
import biotite.structure as struc
import biotite.structure.io as strucio
import biotite.structure.io.npz as npz
import biotite.structure.io.pdbx as pdbx
from ..util import data_dir


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "*.npz"))
)
def test_array_conversion(path):
    npz_file = npz.NpzFile.read(path)
    array1 = npz_file.get_structure()
    npz_file = npz.NpzFile()
    npz_file.set_structure(array1)
    array2 = npz_file.get_structure()
    assert array1 == array2


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "*.npz"))
)
def test_pdbx_consistency(path):
    cif_path = splitext(path)[0] + ".cif"
    array1 = strucio.load_structure(path)
    array2 = strucio.load_structure(cif_path)
    if array2.box is not None:
        assert np.allclose(array1.box, array2.box)
    assert array1.bonds == array2.bonds
    for category in array1.get_annotation_categories():
        assert array1.get_annotation(category).tolist() == \
               array2.get_annotation(category).tolist()
    assert array1.coord.tolist() == array2.coord.tolist()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import datetime
import itertools
import tempfile
import numpy as np
from requests.exceptions import ConnectionError
import pytest
import biotite.database.rcsb as rcsb
import biotite.structure.io.pdb as pdb
import biotite.structure.io.pdbx as pdbx
import biotite.structure.io.mmtf as mmtf
import biotite.sequence.io.fasta as fasta
import biotite.sequence.align as align
from biotite.database import RequestError
from ..util import cannot_connect_to, data_dir


RCSB_URL = "https://www.rcsb.org/"


@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
@pytest.mark.parametrize(
    "format, as_file_like",
    itertools.product(["pdb", "cif", "mmtf", "fasta"], [False, True])
)
def test_fetch(format, as_file_like):
    path = None if as_file_like else tempfile.gettempdir()
    file_path_or_obj = rcsb.fetch("1l2y", format, path, overwrite=True)
    if format == "pdb":
        file = pdb.PDBFile.read(file_path_or_obj)
        pdb.get_structure(file)
    elif format == "pdbx":
        file = pdbx.PDBxFile.read(file_path_or_obj)
        pdbx.get_structure(file)
    elif format == "mmtf":
        file = mmtf.MMTFFile.read(file_path_or_obj)
        mmtf.get_structure(file)
    elif format == "fasta":
        file = fasta.FastaFile.read(file_path_or_obj)
        # Test if the file contains any sequences
        assert len(fasta.get_sequences(file)) > 0


@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
@pytest.mark.parametrize("format", ["pdb", "cif", "mmtf", "fasta"])
def test_fetch_invalid(format):
    with pytest.raises(RequestError):
        file = rcsb.fetch(
            "xxxx", format, tempfile.gettempdir(), overwrite=True
        )


def test_search_basic():
    query = rcsb.BasicQuery("tc5b")
    assert rcsb.search(query) == ["1L2Y"]
    assert rcsb.count(query) == 1


@pytest.mark.parametrize(
    "field, molecular_definition, params, ref_ids",
    [
        (
            "pdbx_serial_crystallography_sample_delivery_injection.preparation",
            False,
            {},
            ["6IG7", "6IG6", "7JRI", "7JR5"]
        ),
        (
            "audit_author.name",
            False,
            {"is_in": ["Neidigh, J.W."]},
            ["1JRJ", "1L2Y", "2O3P", "2O63", "2O64", "2O65"]
        ),
        (
            "rcsb_entity_source_organism.rcsb_gene_name.value",
            False,
            {"exact_match": "lacA"},
            ["5JUV", "1KQA", "1KRV", "1KRU", "1KRR", "1TG7", "1XC6", "3U7V",
             "4IUG", "4LFK", "4LFL", "4LFM", "4LFN", "5IFP", "5IFT", "5IHR",
             "4DUW", "5MGD", "5MGC"]
        ),
        (
            "struct.title",
            False,
            {"contains_words": "tc5b"},
            ["1L2Y"]
        ),
        (
            "reflns.d_resolution_high",
            False,
            {"less_or_equal": 0.6},
            ["1EJG", "1I0T", "3NIR", "3P4J", "5D8V", "5NW3", "4JLJ", "2GLT",
             "7ATG"]
        ),
        (
            "rcsb_entry_info.deposited_model_count",
            False,
            {"range_closed": (60, 61)},
            ["1BBO", "1GB1", "1O5P", "1XU6", "2LUM", "2NO8"]
        ),
        (
            "rcsb_id",
            True,
            {"exact_match": "AIN"},
            ["1OXR", "1TGM", "3IAZ", "3GCL", "6MQF", "2QQT", "4NSB"]
        ),
    ]
)
@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
def test_search_field(field, molecular_definition, params, ref_ids):
    query = rcsb.FieldQuery(
        field, molecular_definition, **params
    )
    test_ids = rcsb.search(query)
    test_count = rcsb.count(query)

    assert set(test_ids) == set(ref_ids)
    assert test_count == len(ref_ids)


@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
def test_search_sequence():
    IDENTIY_CUTOFF = 0.9
    pdbx_file = pdbx.PDBxFile.read(join(data_dir("structure"), "1l2y.cif"))
    ref_sequence = pdbx.get_sequence(pdbx_file)[0]
    query = rcsb.SequenceQuery(
        ref_sequence, "protein", min_identity=IDENTIY_CUTOFF
    )
    test_ids = rcsb.search(query)

    for id in test_ids:
        fasta_file = fasta.FastaFile.read(rcsb.fetch(id, "fasta"))
        test_sequence = fasta.get_sequence(fasta_file)
        matrix = align.SubstitutionMatrix.std_protein_matrix()
        alignment = align.align_optimal(
            ref_sequence, test_sequence, matrix, terminal_penalty=False
        )[0]
        identity = align.get_sequence_identity(alignment, mode="shortest")
        assert identity >= IDENTIY_CUTOFF


@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
def test_search_structure():
    query = rcsb.StructureQuery("1L2Y", chain="A")
    test_ids = rcsb.search(query)
    assert "1L2Y" in test_ids


@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
def test_search_motif():
    # motif is taken from official RCSB search API tutorial
    MOTIF = "C-x(2,4)-C-x(3)-[LIVMFYWC]-x(8)-H-x(3,5)-H."
    query = rcsb.MotifQuery(MOTIF, "prosite", "protein")
    test_count = rcsb.count(query)
    assert test_count == pytest.approx(503, rel=0.1)


@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
def test_search_composite():
    query1 = rcsb.FieldQuery(
        "rcsb_entity_host_organism.scientific_name",
        exact_match="Homo sapiens"
    )
    query2 = rcsb.FieldQuery(
        "exptl.method",
        exact_match="SOLUTION NMR"
    )
    ids_1 = set(rcsb.search(query1))
    ids_2 = set(rcsb.search(query2))
    ids_or = set(rcsb.search(query1 | query2))
    ids_and = set(rcsb.search(query1 & query2))

    assert ids_or  == ids_1 | ids_2
    assert ids_and == ids_1 & ids_2


@pytest.mark.parametrize(
    "return_type, expected",
    [
        ("entry",              ["1L2Y"]  ),
        ("assembly",           ["1L2Y-1"]),
        ("polymer_entity",     ["1L2Y_1"]),
        ("non_polymer_entity", []        ),
        ("polymer_instance",   ["1L2Y.A"]),
    ]
)
@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
def test_search_return_type(return_type, expected):
    query = rcsb.BasicQuery("tc5b")
    assert rcsb.search(query, return_type) == expected
    assert rcsb.count(query, return_type) == len(expected)


def test_search_empty():
    query = rcsb.BasicQuery("This will not match any ID")
    assert rcsb.search(query) == []
    assert rcsb.count(query) == 0


@pytest.mark.parametrize(
    "field, params",
    [
        (
            "invalid.field",
            {"exact_match": "Some Value"}
        ),
        (
            "exptl.method",
            {"less": 5}
        )
    ]
)
@pytest.mark.skipif(
    cannot_connect_to(RCSB_URL),
    reason="RCSB PDB is not available"
)
def test_search_invalid(field, params):
    invalid_query = rcsb.FieldQuery(field, **params)
    with pytest.raises(RequestError, match="400"):
        rcsb.search(invalid_query)
    with pytest.raises(RequestError, match="400"):
        rcsb.count(invalid_query)



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions for base pair identification.
"""

__name__ = "biotite.structure"
__author__ = "Tom David Müller"
__all__ = ["base_pairs", "map_nucleotide", "base_stacking", "base_pairs_edge",
           "Edge", "base_pairs_glycosidic_bond", "GlycosidicBond"]

import numpy as np
import warnings
from enum import IntEnum
from .atoms import Atom, array
from .superimpose import superimpose, superimpose_apply
from .filter import filter_nucleotides
from .celllist import CellList
from .hbond import hbond
from .error import IncompleteStructureWarning, UnexpectedStructureWarning, \
    BadStructureError
from .util import distance, norm_vector
from .residues import get_residue_starts_for, get_residue_masks
from .info.standardize import standardize_order
from .compare import rmsd


def _get_std_adenine():
    """
    Get standard base variables for adenine.

    Returns
    -------
    standard_base : AtomArray
        Standard coordinates nomenclature of the adenine base as
        :class:`AtomArray` with nomenclature of PDB File Format V3
    coordinates : tuple (ndarray, ndarray, ndarray, dtype=float)
        :class:`ndarray` containing the center according to the SCHNaP-
        paper referenced in the function ``base_pairs``,
        :class:`ndarray` containing the coordinates of the pyrimidine
        ring center, :class:`ndarray` containing the coordinates of the
        imidazole ring center
    """
    atom1 =  Atom([-1.291, 4.498, 0.000], atom_name="N9",  res_name="A")
    atom2 =  Atom([0.024, 4.897, 0.000],  atom_name="C8",  res_name="A")
    atom3 =  Atom([0.877, 3.902, 0.000],  atom_name="N7",  res_name="A")
    atom4 =  Atom([0.071, 2.771, 0.000],  atom_name="C5",  res_name="A")
    atom5 =  Atom([0.369, 1.398, 0.000],  atom_name="C6",  res_name="A")
    atom6 =  Atom([1.611, 0.909, 0.000],  atom_name="N6",  res_name="A")
    atom7 =  Atom([-0.668, 0.532, 0.000], atom_name="N1",  res_name="A")
    atom8 =  Atom([-1.912, 1.023, 0.000], atom_name="C2",  res_name="A")
    atom9 =  Atom([-2.320, 2.290, 0.000], atom_name="N3",  res_name="A")
    atom10 = Atom([-1.267, 3.124, 0.000], atom_name="C4",  res_name="A")
    adenine = array(
        [atom1, atom2, atom3, atom4, atom5, atom6, atom7, atom8,
         atom9, atom10]
    )

    # Get the midpoint between the N1 and C4 atoms
    midpoint = np.mean([atom7.coord, atom10.coord], axis=-2)
    # Calculate the coordinates of the aromatic ring centers
    pyrimidine_center = np.mean(
        [atom4.coord, atom5.coord, atom7.coord,
         atom8.coord, atom9.coord, atom10.coord], axis=-2
    )
    imidazole_center = np.mean(
        [atom1.coord, atom2.coord, atom3.coord,
         atom4.coord, atom10.coord], axis=-2
    )

    return adenine, (midpoint, pyrimidine_center, imidazole_center)


def _get_std_cytosine():
    """
    Get standard base variables for cytosine.

    Returns
    -------
   standard_base : AtomArray
        Standard coordinates nomenclature of the cytosine base as
        :class:`AtomArray` with nomenclature of PDB File Format V3
    coordinates : tuple (ndarray, ndarray, dtype=float)
        :class:`ndarray` containing the center according to the SCHNaP-
        paper referenced in the function ``base_pairs``,
        :class:`ndarray` containing the coordinates of the pyrimidine
        ring center
    """
    atom1 = Atom([-1.285, 4.542, 0.000], atom_name="N1",  res_name="C")
    atom2 = Atom([-1.472, 3.158, 0.000], atom_name="C2",  res_name="C")
    atom3 = Atom([-2.628, 2.709, 0.000], atom_name="O2",  res_name="C")
    atom4 = Atom([-0.391, 2.344, 0.000], atom_name="N3",  res_name="C")
    atom5 = Atom([0.837, 2.868, 0.000],  atom_name="C4",  res_name="C")
    atom6 = Atom([1.875, 2.027, 0.000],  atom_name="N4",  res_name="C")
    atom7 = Atom([1.056, 4.275, 0.000],  atom_name="C5",  res_name="C")
    atom8 = Atom([-0.023, 5.068, 0.000], atom_name="C6",  res_name="C")
    cytosine = array(
        [atom1, atom2, atom3, atom4, atom5, atom6, atom7, atom8]
    )

    # Get the midpoint between the N3 and C6 atoms
    midpoint = np.mean([atom4.coord, atom8.coord], axis=-2)
    # Calculate the coordinates of the aromatic ring center
    pyrimidine_center = np.mean(
        [atom1.coord, atom2.coord, atom4.coord,
         atom5.coord, atom7.coord, atom8.coord], axis=-2
    )

    return cytosine, (midpoint, pyrimidine_center)


def _get_std_guanine():
    """
    Get standard base variables for guanine.

    Returns
    -------
    standard_base : AtomArray
        Standard coordinates nomenclature of the guanine base as
        :class:`AtomArray` with nomenclature of PDB File Format V3
    coordinates : tuple (ndarray, ndarray, ndarray, dtype=float)
        :class:`ndarray` containing the center according to the SCHNaP-
        paper referenced in the function ''base_pairs'',
        :class:`ndarray` containing the coordinates of the pyrimidine
        ring center, :class:`ndarray` containing the coordinates of the
        imidazole ring center
    """
    atom1 =  Atom([-1.289, 4.551, 0.000],  atom_name="N9",  res_name="G")
    atom2 =  Atom([0.023, 4.962, 0.000],   atom_name="C8",  res_name="G")
    atom3 =  Atom([0.870, 3.969, 0.000],   atom_name="N7",  res_name="G")
    atom4 =  Atom([0.071, 2.833, 0.000],   atom_name="C5",  res_name="G")
    atom5 =  Atom([0.424, 1.460, 0.000],   atom_name="C6",  res_name="G")
    atom6 =  Atom([1.554, 0.955, 0.000],   atom_name="O6",  res_name="G")
    atom7 =  Atom([-0.700, 0.641, 0.000],  atom_name="N1",  res_name="G")
    atom8 =  Atom([-1.999, 1.087, 0.000],  atom_name="C2",  res_name="G")
    atom9 =  Atom([-2.949, 0.139, -0.001], atom_name="N2",  res_name="G")
    atom10 = Atom([-2.342, 2.364, 0.001],  atom_name="N3",  res_name="G")
    atom11 = Atom([-1.265, 3.177, 0.000],  atom_name="C4",  res_name="G")
    guanine = array(
        [atom1, atom2, atom3, atom4, atom5, atom6, atom7, atom8,
         atom9, atom10, atom11]
    )

    # Get the midpoint between the N1 and C4 atoms
    midpoint = np.mean([atom7.coord, atom11.coord], axis=-2)
    # Calculate the coordinates of the aromatic ring centers
    pyrimidine_center = np.mean(
        [atom4.coord, atom5.coord, atom7.coord,
         atom8.coord, atom10.coord, atom11.coord], axis=-2
    )
    imidazole_center = np.mean(
        [atom1.coord, atom2.coord, atom3.coord,
         atom4.coord, atom11.coord], axis=-2
    )

    return guanine, (midpoint, pyrimidine_center, imidazole_center)


def _get_std_thymine():
    """
    Get standard base variables for thymine.

    Returns
    -------
    standard_base : AtomArray
        Standard coordinates nomenclature of the thymine base as
        :class:`AtomArray` with nomenclature of PDB File Format V3
    coordinates : tuple (ndarray, ndarray, dtype=float)
        :class:`ndarray` containing the center according to the SCHNaP-
        paper referenced in the function ``base_pairs``,
        :class:`ndarray` containing the coordinates of the pyrimidine
        ring center
    """
    atom1 = Atom([-1.284, 4.500, 0.000], atom_name="N1",  res_name="T")
    atom2 = Atom([-1.462, 3.135, 0.000], atom_name="C2",  res_name="T")
    atom3 = Atom([-2.562, 2.608, 0.000], atom_name="O2",  res_name="T")
    atom4 = Atom([-0.298, 2.407, 0.000], atom_name="N3",  res_name="T")
    atom5 = Atom([0.994, 2.897, 0.000],  atom_name="C4",  res_name="T")
    atom6 = Atom([1.944, 2.119, 0.000],  atom_name="O4",  res_name="T")
    atom7 = Atom([1.106, 4.338, 0.000],  atom_name="C5",  res_name="T")
    atom8 = Atom([2.466, 4.961, 0.001],  atom_name="C7", res_name="T")
    atom9 = Atom([-0.024, 5.057, 0.000], atom_name="C6",  res_name="T")
    thymine = array(
        [atom1, atom2, atom3, atom4, atom5, atom6, atom7, atom8, atom9]
    )

    # Get the midpoint between the N3 and C6 atoms
    midpoint = np.mean([atom4.coord, atom9.coord], axis=-2)
    # Calculate the coordinates of the aromatic ring center
    pyrimidine_center = np.mean(
        [atom1.coord, atom2.coord, atom4.coord,
         atom5.coord, atom7.coord, atom9.coord], axis=-2
    )

    return thymine, (midpoint, pyrimidine_center)


def _get_std_uracil():
    """
    Get standard base variables for uracil.

    Returns
    -------
    standard_base : AtomArray
        Standard coordinates nomenclature of the uracil base as
        :class:`AtomArray` with nomenclature of PDB File Format V3
    coordinates : tuple (ndarray, ndarray, dtype=float)
        :class:`ndarray` containing the center according to the SCHNaP-
        paper referenced in the function ``base_pairs``,
        :class:`ndarray` containing the coordinates of the pyrimidine
        ring center
    """
    atom1 = Atom([-1.284, 4.500, 0.000], atom_name="N1",  res_name="U")
    atom2 = Atom([-1.462, 3.131, 0.000], atom_name="C2",  res_name="U")
    atom3 = Atom([-2.563, 2.608, 0.000], atom_name="O2",  res_name="U")
    atom4 = Atom([-0.302, 2.397, 0.000], atom_name="N3",  res_name="U")
    atom5 = Atom([0.989, 2.884, 0.000],  atom_name="C4",  res_name="U")
    atom6 = Atom([1.935, 2.094, -0.001], atom_name="O4",  res_name="U")
    atom7 = Atom([1.089, 4.311, 0.000],  atom_name="C5",  res_name="U")
    atom8 = Atom([-0.024, 5.053, 0.000], atom_name="C6",  res_name="U")
    uracil = array(
        [atom1, atom2, atom3, atom4, atom5, atom6, atom7, atom8]
    )

    # Get the midpoint between the N3 and C6 atoms
    midpoint = np.mean([atom4.coord, atom8.coord], axis=-2)
    # Calculate the coordinates of the aromatic ring center
    pyrimidine_center = np.mean(
        [atom1.coord, atom2.coord, atom4.coord,
         atom5.coord, atom7.coord, atom8.coord], axis=-2
    )

    return uracil, (midpoint, pyrimidine_center)


_STD_ADENINE, _STD_ADENINE_RING_CENTERS  = _get_std_adenine()
_STD_CYTOSINE, _STD_CYTOSINE_RING_CENTERS = _get_std_cytosine()
_STD_GUANINE, _STD_GUANINE_RING_CENTERS = _get_std_guanine()
_STD_THYMINE, _STD_THYMINE_RING_CENTERS = _get_std_thymine()
_STD_URACIL, _STD_URACIL_RING_CENTERS = _get_std_uracil()

_ADENINE_CONTAINING_NUCLEOTIDES = ["A", "DA"]
_THYMINE_CONTAINING_NUCLEOTIDES = ["T", "DT"]
_CYTOSINE_CONTAINING_NUCLEOTIDES = ["C", "DC"]
_GUANINE_CONTAINING_NUCLEOTIDES = ["G", "DG"]
_URACIL_CONTAINING_NUCLEOTIDES = ["U", "DU"]
_REFERENCE_NUCLEOTIDE_NAMES = (
    _ADENINE_CONTAINING_NUCLEOTIDES +
    _THYMINE_CONTAINING_NUCLEOTIDES +
    _CYTOSINE_CONTAINING_NUCLEOTIDES +
    _GUANINE_CONTAINING_NUCLEOTIDES +
    _URACIL_CONTAINING_NUCLEOTIDES
)

# Atoms that are part of respective base edges according to the
# Leontis-Westhof nomenclature
_WATSON_CRICK_EDGE = {
    "A" : ["N6", "N1"],
    "G" : ["O6", "N1", "N2"],
    "U" : ["O4", "N3", "O2"],
    "T" : ["O4", "N3", "O2"],
    "C" : ["N4", "N3", "O2"]
}
_HOOGSTEEN_EDGE = {
    "A" : ["N6", "N7"],
    "G" : ["O6", "N7"],
    "U" : ["O4"],
    "T" : ["O4"],
    "C" : ["N4"]
}
_SUGAR_EDGE = {
    "A" : ["N3", "O2'"],
    "G" : ["N2", "N3", "O2'"],
    "U" : ["O2", "O2'"],
    "T" : ["O2", "O2'"],
    "C" : ["O2", "O2'"]
}
_EDGES = [_WATSON_CRICK_EDGE, _HOOGSTEEN_EDGE, _SUGAR_EDGE]


class Edge(IntEnum):
    """
    This enum type represents the interacting edge for a given base.
    """
    INVALID = 0,
    WATSON_CRICK = 1,
    HOOGSTEEN = 2,
    SUGAR = 3


class GlycosidicBond(IntEnum):
    """
    This enum type represents the relative glycosidic bond orientation
    for a given base pair.
    """
    INVALID = 0
    CIS = 1,
    TRANS = 2,


def base_pairs_edge(atom_array, base_pairs):
    """
    Get the interacting edges for given base pairs in an
    :class:`AtomArray` according to the Leontis-Westhof nomenclature.
    :footcite:`Leontis2001`

    The :class:`AtomArray` must contain hydrogens as it relies on
    :func:`hbond()`.

    Parameters
    ----------
    atom_array : AtomArray
        The :class:`AtomArray` containing the bases.
    base_pairs : ndarray, dtype=int, shape=(n,2)
        Each row is equivalent to one base pair and contains the first
        indices of the residues corresponding to each base. The
        structure of the ``ndarray`` is the same as the output of
        :func:`base_pairs()`.

    Returns
    -------
    results : ndarray, dtype=uint8, shape=(n,2)
        The ``ndarray`` has the same dimensions as ``base_pairs``. Each
        cell corresponds to the interacting edge of the referenced base
        in ``base_pairs``. The edge type is stored as integer that is
        interpreted as member of the the :class:`Edge` enum.

    See Also
    --------
    base_pairs
    base_pairs_glycosidic_bond

    Notes
    -----
    If a base is not a canonical base (``A``, ``C``, ``G``, ``T``,
    ``U``) or no hydrogen bonds are found between the bases that conform
    to the interacting edges described by Leontis and Westhof, 0 is
    returned (corresponding to ``Edge.INVALID``).

    The edge returned always corresponds to the edge with the most
    hydrogen bonding interactions.

    Examples
    --------
    Compute the interacting base edges for the dna helix with the PDB
    id 1QXB:

    >>> from os.path import join
    >>> dna_helix = load_structure(
    ...     join(path_to_structures, "base_pairs", "1qxb.cif")
    ... )
    >>> basepairs = base_pairs(dna_helix)
    >>> interacting_edges = base_pairs_edge(dna_helix, basepairs)
    >>> print(interacting_edges)
    [[1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]
     [1 1]]

    The resulting integers can be interpreted as :class:`Edge` ``Enum``:

    >>> for interaction in interacting_edges:
    ...     print(Edge(interaction[0]), Edge(interaction[1]))
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK
    Edge.WATSON_CRICK Edge.WATSON_CRICK

    References
    ----------
    
    .. footbibliography::
    """
    # Result-``ndarray`` matches the dimensions of the input array
    results = np.zeros_like(base_pairs, dtype='uint8')

    # Get the residue masks for each residue
    base_pairs_masks = get_residue_masks(atom_array, base_pairs.flatten())

    # Group every two masks together for easy iteration (each 'row' is
    # respective to a row in ``base_pairs``)
    base_pairs_masks = base_pairs_masks.reshape(
        (base_pairs.shape[0], 2, atom_array.shape[0])
    )

    for i, base_masks in enumerate(base_pairs_masks):
        # Get the absolute atom count for each edge
        base_edges = _get_edge_matrix(atom_array, base_masks)

        # Classify the base edges based on the highest number of
        # matching hydrogen bonded atoms
        for j, base in enumerate(base_edges):
            if np.max(base) != 0:
                results[i, j] = np.argmax(base) + 1
    return results


def _get_edge_matrix(atom_array, base_masks):
    """
    Get the number of atoms interacting for each edge as a matrix, where
    each row corresponds to a base and each column to the number of
    Watson-Crick-, Hoogsteen- and Sugar-edge interactions respectively.

    Parameters
    ----------
    atom_array : AtomArray
        The :class:`AtomArray` containing the bases.
    base_masks : ndarray, dtype=bool, shape=(2,n)
        Boolean masks for the interacting bases

    Returns
    -------
    matrix : ndarray, dtype=int, shape=(2,3)
        The edge matrix.
    """
    # Get the hydrogen bonds between the residues
    hbonds = hbond(atom_array, base_masks[0], base_masks[1])
    if len(hbonds) == 0:
        raise BadStructureError(
            f"No hydrogen bonds between nucleotides with residue start "
            f"indices {np.argmax(base_masks[0])} and "
            f"{np.argmax(base_masks[1])}"
        )
    # filter out donor/acceptor heteroatoms and flatten for easy
    # iteration
    hbonds = hbonds[:, (0,2)].flatten()

    # ``ndarray`` with one row for each base and the number of
    # bonded edge heteroatoms as in ``_edge`` as columns
    matrix = np.zeros((2, 3), dtype='int32')

    # Iterate through the atoms and corresponding atoms indices
    # that are part of the hydrogen bonds
    for atom, atom_index in zip(atom_array[hbonds], hbonds):

        if atom.res_name not in _REFERENCE_NUCLEOTIDE_NAMES:
            continue

        # Iterate over the edge types
        for edge_type_index, edge_type in enumerate(_EDGES):
            # Iterate over the two base masks
            for base_index, base_mask in enumerate(base_masks):
                # If a donor/acceptor atom name matches a name in
                # the corresponding edge list increase the tally
                if (base_mask[atom_index] and
                    atom.atom_name in edge_type[atom.res_name[-1]]):
                    matrix[base_index, edge_type_index] += 1
    return matrix


def base_pairs_glycosidic_bond(atom_array, base_pairs):
    """
    Calculate the glycosidic bond orientation for given base pairs in an
    :class:`AtomArray` according to the Leontis-Westhof nomenclature.
    :footcite:`Leontis2001`

    Parameters
    ----------
    atom_array : AtomArray
        The :class:`AtomArray` containing the bases.
    base_pairs : ndarray, dtype=int, shape=(n,2)
        Each row is equivalent to one base pair and contains the first
        indices of the residues corresponding to each base. The
        structure of the ``ndarray`` is the same as the output of
        :func:`base_pairs()`.

    Returns
    -------
    results : ndarray, dtype=edge, shape=(n,)
        The ``ndarray`` has the same dimensions as ``base_pairs``. Each
        cell corresponds to the interacting edge of the referenced base
        in ``base_pairs``.
        Each row is equivalent to the respective base pair. The
        glycosidic bond orientation is stored as integer that is
        interpreted as member of the the :class:`GlycosidicBond` class.

    See Also
    --------
    base_pairs
    base_pairs_edge
    GlycosidicBond

    Notes
    -----
    The orientation is found using the geometric centers of the bases
    and the glycosidic bonds as described in :footcite:`Yang2003`.

    Examples
    --------
    Compute the glycosidic bond orientations for the dna helix with the
    PDB ID 1QXB:

    >>> from os.path import join
    >>> dna_helix = load_structure(
    ...     join(path_to_structures, "base_pairs", "1qxb.cif")
    ... )
    >>> basepairs = base_pairs(dna_helix)
    >>> orientations = base_pairs_glycosidic_bond(dna_helix, basepairs)
    >>> print(orientations)
    [1 1 1 1 1 1 1 1 1 1 1 1]

    The resulting integers can be interpreted as :class:`GlycosidicBond`
    ``Enum``:

    >>> for orientation in orientations:
    ...     print(GlycosidicBond(orientation))
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS
    GlycosidicBond.CIS

    References
    ----------
    
    .. footbibliography::
    """
    results = np.zeros(len(base_pairs), dtype='uint8')

    # Get the residue masks for each residue
    base_pairs_masks = get_residue_masks(atom_array, base_pairs.flatten())

    # Group every two masks together for easy iteration (each 'row' is
    # respective to a row in ``base_pairs``)
    base_pairs_masks = base_pairs_masks.reshape(
        (base_pairs.shape[0], 2, atom_array.shape[0])
    )

    for i, pair_masks in enumerate(base_pairs_masks):

        # position vectors of each bases geometric center
        geometric_centers = np.zeros((2, 3))
        # direction vectors of the glycosidic bonds
        glycosidic_bonds = np.zeros((2, 3))

        for base_index, base_mask in enumerate(pair_masks):
            base = atom_array[base_mask]
            ring_center = _match_base(base, 3)[3:]

            # For Purines the glycosidic bond is between the C1' and the
            # N9 atoms, for pyrimidines it is between the C1' atom and
            # the N1 atom
            if (base.res_name[0] in _ADENINE_CONTAINING_NUCLEOTIDES or
                base.res_name[0] in _GUANINE_CONTAINING_NUCLEOTIDES):

                geometric_centers[base_index] = (
                    (ring_center[0] + ring_center[1]) / 2
                )
                base_atom = base[base.atom_name == "N9"][0]

            elif (base.res_name[0] in _THYMINE_CONTAINING_NUCLEOTIDES or
                base.res_name[0] in _URACIL_CONTAINING_NUCLEOTIDES or
                base.res_name[0] in _CYTOSINE_CONTAINING_NUCLEOTIDES):

                geometric_centers[base_index] = ring_center[0]
                base_atom = base[base.atom_name == "N1"][0]

            else:

                results[i] = GlycosidicBond.INVALID
                break

            sugar_atom = base[base.atom_name == "C1'"][0]

            # Calculate the glycosidic bond direction vector
            glycosidic_bonds[base_index] = sugar_atom.coord - base_atom.coord

        # if the bond is not invalid compute the orientation
        else:
            # Calculate the direction vector between the geometric centers
            geometric_centers_dir = geometric_centers[1] - geometric_centers[0]

            # Check the orientation of the glycosidic bonds
            if np.dot(
                np.cross(geometric_centers_dir, glycosidic_bonds[0]),
                np.cross(geometric_centers_dir, glycosidic_bonds[1])
            ) < 0:

                results[i] = GlycosidicBond.TRANS

            else:

                results[i] = GlycosidicBond.CIS

    return results


def base_stacking(atom_array, min_atoms_per_base=3):
    """
    Find pi-stacking interactions between aromatic rings
    in nucleic acids.

    The presence of base stacking is assumed if the following criteria
    are met :footcite:`Gabb1996`:

    (i) Distance between aromatic ring centers <=4.5 Å

    (ii) Angle between the ring normal vectors <=23°

    (iii) Angle between normalized distance vector between two ring
          centers and both bases' normal vectors <=40°

    Parameters
    ----------
    atom_array : AtomArray
        The :class:`AtomArray` to find stacked bases in.
    min_atoms_per_base : integer, optional (default: 3)
        The number of atoms a nucleotides' base must have to be
        considered a candidate for a stacking interaction.

    Returns
    -------
    stacked_bases : ndarray, dtype=int, shape=(n,2)
        Each row is equivalent to one pair of stacked bases and
        contains the indices to the first atom for each one of both
        paired residues.

    Notes
    -----
    Please note that ring normal vectors are assumed to be equal to the
    base normal vectors.

    Examples
    --------
    Compute the stacking interactions for a DNA-double-helix (PDB ID
    1BNA):

    >>> from os.path import join
    >>> dna_helix = load_structure(
    ...     join(path_to_structures, "base_pairs", "1bna.pdb")
    ... )
    >>> stacking_interactions = base_stacking(dna_helix)
    >>> print(dna_helix[stacking_interactions].res_id)
    [[ 1  2]
     [ 2  3]
     [ 3  4]
     [ 4  5]
     [ 5  6]
     [ 6  7]
     [ 7  8]
     [ 8  9]
     [ 9 10]
     [11 12]
     [14 15]
     [15 16]
     [16 17]
     [17 18]
     [18 19]
     [19 20]
     [20 21]
     [21 22]
     [22 23]
     [23 24]]

    References
    ----------
    
    .. footbibliography::
    """
    # Get the stacking candidates according to a cutoff distance, where
    # each base is identified as the first index of its respective
    # residue.
    # The diameter from the C1'-sugar-atom across a purine base is ~5Å
    # and the distance between the base centers can be at most 4.5Å.
    # Thus, accounting for buffer, a cutoff of 15Å between the
    # nucleotides' C1'-atoms was chosen.
    c1_mask = filter_nucleotides(atom_array) & (atom_array.atom_name == "C1'")
    stacking_candidates, _ = _get_proximate_residues(atom_array, c1_mask, 15)

    # Contains the plausible pairs of stacked bases
    stacked_bases = []

    # Get the residue masks for each residue
    base_masks = get_residue_masks(atom_array, stacking_candidates.flatten())

    # Group every two masks together for easy iteration (each 'row' is
    # respective to a row in ``stacking_candidates``)
    base_masks = base_masks.reshape(
        (stacking_candidates.shape[0], 2, atom_array.shape[0])
    )

    for (base1_index, base2_index), (base1_mask, base2_mask) in zip(
        stacking_candidates, base_masks
    ):
        bases = (atom_array[base1_mask], atom_array[base2_mask])

        # A list containing ndarray for each base with transformed
        # vectors from the standard base reference frame to the
        # structures' coordinates. The layout is as follows:
        #
        # [Origin coordinates]
        # [Base normal vector]
        # [SCHNAaP origin coordinates]
        # [Aromatic Ring Center coordinates]
        transformed_std_vectors = [None] * 2

        # Generate the data necessary for analysis of each base.
        for i in range(2):
            base_tuple = _match_base(bases[i], min_atoms_per_base)

            if(base_tuple is None):
                break

            transformed_std_vectors[i] = base_tuple

        normal_vectors = np.vstack((transformed_std_vectors[0][1],
                                    transformed_std_vectors[1][1]))
        aromatic_ring_centers = [transformed_std_vectors[0][3:],
                                        transformed_std_vectors[1][3:]]

        # Check if the base pairs are stacked.
        stacked = _check_base_stacking(aromatic_ring_centers, normal_vectors)

        # If a stacking interaction is found, append the first indices
        # of the bases´'residues to the output.
        if stacked:
            stacked_bases.append((base1_index, base2_index))

    return np.array(stacked_bases)


def base_pairs(atom_array, min_atoms_per_base = 3, unique = True):
    """
    Use DSSR criteria to find the base pairs in an :class:`AtomArray`.

    The algorithm is able to identify canonical and non-canonical
    base pairs. between the 5 common bases Adenine, Guanine, Thymine,
    Cytosine, and Uracil bound to Deoxyribose and Ribose.
    Each Base is mapped to the 5 common bases Adenine, Guanine, Thymine,
    Cytosine, and Uracil in a standard reference frame described in
    :footcite:`Olson2001` using :func:`map_nucleotide()`.

    The DSSR Criteria are as follows :footcite:`Lu2015`:

    (i) Distance between base origins <=15 Å

    (ii) Vertical separation between the base planes <=2.5 Å

    (iii) Angle between the base normal vectors <=65°

    (iv) Absence of stacking between the two bases

    (v) Presence of at least one hydrogen bond involving a base atom

    Parameters
    ----------
    atom_array : AtomArray
        The :class:`AtomArray` to find base pairs in.
    min_atoms_per_base : integer, optional (default: 3)
        The number of atoms a nucleotides' base must have to be
        considered a candidate for a base pair.
    unique : bool, optional (default: True)
        If ``True``, each base is assumed to be only paired with one
        other base. If multiple pairings are plausible, the pairing with
        the most hydrogen bonds is selected.

    Returns
    -------
    basepairs : ndarray, dtype=int, shape=(n,2)
        Each row is equivalent to one base pair and contains the first
        indices of the residues corresponding to each base.

    Notes
    -----
    The bases from the standard reference frame described in
    :footcite:`Olson2001` were modified such that only the base atoms
    are implemented.
    Sugar atoms (specifically C1') were disregarded, as nucleosides such
    as PSU do not posess the usual N-glycosidic linkage, thus leading to
    inaccurate results.

    The vertical separation is implemented as the scalar
    projection of the distance vectors between the base origins
    according to :footcite:`Lu1997` onto the averaged base normal
    vectors.

    The presence of base stacking is assumed if the following criteria
    are met :footcite:`Gabb1996`:

    (i) Distance between aromatic ring centers <=4.5 Å

    (ii) Angle between the ring normal vectors <=23°

    (iii) Angle between normalized distance vector between two ring
          centers and both bases' normal vectors <=40°

    Please note that ring normal vectors are assumed to be equal to the
    base normal vectors.

    For structures without hydrogens the accuracy of the algorithm is
    limited as the hydrogen bonds can be only checked be checked for
    plausibility.
    A hydrogen bond is considered as plausible if a cutoff of 3.6 Å
    between N/O atom pairs is met. 3.6Å was chosen as hydrogen bonds are
    typically 1.5-2.5Å in length. N-H and O-H bonds have a length of
    1.00Å and 0.96Å respectively. Thus, including some buffer, a 3.6Å
    cutoff should cover all hydrogen bonds.

    Examples
    --------
    Compute the base pairs for the structure with the PDB ID 1QXB:

    >>> from os.path import join
    >>> dna_helix = load_structure(
    ...     join(path_to_structures, "base_pairs", "1qxb.cif")
    ... )
    >>> basepairs = base_pairs(dna_helix)
    >>> print(dna_helix[basepairs].res_name)
    [['DC' 'DG']
     ['DG' 'DC']
     ['DC' 'DG']
     ['DG' 'DC']
     ['DA' 'DT']
     ['DA' 'DT']
     ['DT' 'DA']
     ['DT' 'DA']
     ['DC' 'DG']
     ['DG' 'DC']
     ['DC' 'DG']
     ['DG' 'DC']]

    References
    ----------
    
    .. footbibliography::
    """

    # Get the nucleotides for the given atom_array
    nucleotides_boolean = filter_nucleotides(atom_array)

    # Disregard the phosphate-backbone
    non_phosphate_boolean = (
        ~ np.isin(
            atom_array.atom_name,
            ["O5'", "P", "OP1", "OP2", "OP3", "HOP2", "HOP3"]
        )
    )

    # Combine the two boolean masks
    boolean_mask = nucleotides_boolean & non_phosphate_boolean

    # Get only nucleosides
    nucleosides = atom_array[boolean_mask]


    # Get the base pair candidates according to a N/O cutoff distance,
    # where each base is identified as the first index of its respective
    # residue
    n_o_mask = np.isin(nucleosides.element, ["N", "O"])
    basepair_candidates, n_o_matches = _get_proximate_residues(
        nucleosides, n_o_mask, 3.6
    )

    # Contains the plausible base pairs
    basepairs = []
    # Contains the number of hydrogens for each plausible base pair
    basepairs_hbonds = []

    # Get the residue masks for each residue
    base_masks = get_residue_masks(nucleosides, basepair_candidates.flatten())

    # Group every two masks together for easy iteration (each 'row' is
    # respective to a row in ``basepair_candidates``)
    base_masks = base_masks.reshape(
        (basepair_candidates.shape[0], 2, nucleosides.shape[0])
    )

    for (base1_index, base2_index), (base1_mask, base2_mask), n_o_pairs in zip(
        basepair_candidates, base_masks, n_o_matches
    ):
        base1 = nucleosides[base1_mask]
        base2 = nucleosides[base2_mask]

        hbonds =  _check_dssr_criteria(
            (base1, base2), min_atoms_per_base, unique
        )

        # If no hydrogens are present use the number N/O pairs to
        # decide between multiple pairing possibilities.

        if hbonds is None:
            # Each N/O-pair is detected twice. Thus, the number of
            # matches must be divided by two.
            hbonds = n_o_pairs/2
        if hbonds != -1:
            basepairs.append((base1_index, base2_index))
            if unique:
                basepairs_hbonds.append(hbonds)

    basepair_array = np.array(basepairs)

    if unique:
        # Contains all non-unique base pairs that are flagged to be
        # removed
        to_remove = []

        # Get all bases that have non-unique pairing interactions
        base_indices, occurrences = np.unique(basepairs, return_counts=True)
        for base_index, occurrence in zip(base_indices, occurrences):
            if(occurrence > 1):
                # Write the non-unique base pairs to a dictionary as
                # 'index: number of hydrogen bonds'
                remove_candidates = {}
                for i, row in enumerate(
                    np.asarray(basepair_array == base_index)
                ):
                    if(np.any(row)):
                        remove_candidates[i] = basepairs_hbonds[i]
                # Flag all non-unique base pairs for removal except the
                # one that has the most hydrogen bonds
                del remove_candidates[
                    max(remove_candidates, key=remove_candidates.get)
                ]
                to_remove += list(remove_candidates.keys())
        # Remove all flagged base pairs from the output `ndarray`
        basepair_array = np.delete(basepair_array, to_remove, axis=0)

    # Remap values to original atom array
    if len(basepair_array) > 0:
        basepair_array = np.where(boolean_mask)[0][basepair_array]
        for i, row in enumerate(basepair_array):
            basepair_array[i] = get_residue_starts_for(atom_array, row)
    return basepair_array


def _check_dssr_criteria(basepair, min_atoms_per_base, unique):
    """
    Check the DSSR criteria of a potential base pair.

    Parameters
    ----------
    basepair : tuple (AtomArray, AtomArray)
        The two bases to check the criteria for as :class:`AtomArray`.
    min_atoms_per_base : int
        The number of atoms a nucleotides' base must have to be
        considered a candidate for a base pair.
    unique : bool
        If ``True``, the shortest hydrogen bond length between the bases
        is calculated for plausible base pairs.

    Returns
    -------
    satisfied : int
        `> 0` if the base pair satisfies the criteria and `-1`,
        if it does not.
        If unique is ``True``, the number of hydrogen bonds is
        returned for plausible base pairs.
    """

    # A list containing ndarray for each base with transformed
    # vectors from the standard base reference frame to the structures'
    # coordinates. The layout is as follows:
    #
    # [Origin coordinates]
    # [Base normal vector]
    # [SCHNAaP origin coordinates]
    # [Aromatic Ring Center coordinates]
    transformed_std_vectors = [None] * 2

    # Generate the data necessary for analysis of each base.
    for i in range(2):
        transformed_std_vectors[i] = _match_base(
            basepair[i], min_atoms_per_base
        )

        if(transformed_std_vectors[i] is None):
            return -1

    origins = np.vstack((transformed_std_vectors[0][0],
                         transformed_std_vectors[1][0]))
    normal_vectors = np.vstack((transformed_std_vectors[0][1],
                                transformed_std_vectors[1][1]))
    schnaap_origins = np.vstack((transformed_std_vectors[0][2],
                                 transformed_std_vectors[1][2]))
    aromatic_ring_centers = [transformed_std_vectors[0][3:],
                                       transformed_std_vectors[1][3:]]

    # Criterion 1: Distance between orgins <=15 Å
    if not (distance(origins[0], origins[1]) <= 15):
        return -1

    # Criterion 2: Vertical separation <=2.5 Å
    #
    # Average the base normal vectors. If the angle between the vectors
    # is >=90°, flip one vector before averaging
    mean_normal_vector = (
        normal_vectors[0] + (normal_vectors[1] * np.sign(np.dot(
            normal_vectors[0], normal_vectors[1]
        )))
    ) / 2
    norm_vector(mean_normal_vector)
    # Calculate the distance vector between the two SCHNAaP origins
    origin_distance_vector = schnaap_origins[1] - schnaap_origins[0]

    # The scalar projection of the distance vector between the two
    # origins onto the averaged normal vectors is the vertical
    # seperation
    if not abs(np.dot(origin_distance_vector, mean_normal_vector)) <= 2.5:
        return -1

    # Criterion 3: Angle between normal vectors <=65°
    if not (np.arccos(np.dot(normal_vectors[0], normal_vectors[1]))
            >= ((115*np.pi)/180)):
        return -1

    # Criterion 4: Absence of stacking
    if _check_base_stacking(aromatic_ring_centers, normal_vectors):
        return -1

    # Criterion 5: Presence of at least one hydrogen bond
    #
    # Check if both bases came with hydrogens.
    if (("H" in basepair[0].element)
        and ("H" in basepair[1].element)):
        # For Structures that contain hydrogens, check for their
        # presence directly.
        #
        # Generate input atom array for ``hbond``
        potential_basepair = basepair[0] + basepair[1]

        # Get the number of hydrogen bonds
        bonds = len(hbond(
            potential_basepair,
            np.ones_like(potential_basepair, dtype=bool),
            np.ones_like(potential_basepair, dtype=bool)
        ))

        if bonds > 0:
            return bonds
        return -1

    else:
        # If the structure does not contain hydrogens return None
        return None


def _check_base_stacking(aromatic_ring_centers, normal_vectors):
    """
    Check for base stacking between two bases.

    Parameters
    ----------
    aromatic_ring_centers : list [ndarray, ndarray]
        A list with the aromatic ring center coordinates as
        :class:`ndarray`. Each row represents a ring center.
    normal_vectors : ndarray shape=(2, 3)
        The normal vectors of the bases.

    Returns
    -------
    base_stacking : bool
        ``True`` if base stacking is detected and ``False`` if not
    """

    # Contains the normalized distance vectors between ring centers less
    # than 4.5 Å apart.
    normalized_distance_vectors = []

    # Criterion 1: Distance between aromatic ring centers <=4.5 Å
    wrong_distance = True
    for ring_center1 in aromatic_ring_centers[0]:
        for ring_center2 in aromatic_ring_centers[1]:
            if (distance(ring_center1, ring_center2) <= 4.5):
                wrong_distance = False
                normalized_distance_vectors.append(ring_center2 - ring_center1)
                norm_vector(normalized_distance_vectors[-1])
    if wrong_distance:
        return False

    # Criterion 2: Angle between normal vectors or its supplement <=23°
    normal_vectors_angle = np.rad2deg(
        np.arccos(np.dot(normal_vectors[0], normal_vectors[1]))
    )
    if (normal_vectors_angle >= 23) and (normal_vectors_angle <= 157):
        return False

    # Criterion 3: Angle between one normalized distance vector and
    # each of the bases' normal vector or supplement <=40°
    for normal_vector in normal_vectors:
        for normalized_dist_vector in normalized_distance_vectors:
            dist_normal_vector_angle = np.rad2deg(
                np.arccos(np.dot(normal_vector, normalized_dist_vector))
            )
            if ((dist_normal_vector_angle >= 40) and
                (dist_normal_vector_angle <= 120)):
                return False

    return True


def _match_base(nucleotide, min_atoms_per_base):
    """
    Match the nucleotide to a corresponding standard base reference
    frame.

    Parameters
    ----------
    nucleotide : AtomArray
        The nucleotide to be matched to a standard base.
    min_atoms_per_base : integer
        The number of atoms a base must have to be considered a
        candidate for a base pair.

    Returns
    -------
    vectors : ndarray, dtype=float, shape=(n,3)
        Transformed standard vectors, origin coordinates, base normal
        vector, aromatic ring center coordinates.
    """

    # Standard vectors containing the origin and the base normal vectors
    vectors = np.array([[0, 0, 0], [0, 0, 1]], dtype=float)

    # Map the nucleotide to a reference base
    one_letter_code, _ = map_nucleotide(nucleotide, min_atoms_per_base)

    if one_letter_code is None:
        return None

    if (one_letter_code == 'A'):
        std_base = _STD_ADENINE
        std_ring_centers = _STD_ADENINE_RING_CENTERS
    elif (one_letter_code == 'T'):
        std_base = _STD_THYMINE
        std_ring_centers = _STD_THYMINE_RING_CENTERS
    elif (one_letter_code == 'C'):
        std_base = _STD_CYTOSINE
        std_ring_centers = _STD_CYTOSINE_RING_CENTERS
    elif (one_letter_code == 'G'):
        std_base = _STD_GUANINE
        std_ring_centers = _STD_GUANINE_RING_CENTERS
    elif (one_letter_code == 'U'):
        std_base = _STD_URACIL
        std_ring_centers = _STD_URACIL_RING_CENTERS

    # Add the ring centers to the array of vectors to be transformed.
    vectors = np.vstack((vectors, std_ring_centers))

    # Select the matching atoms of the nucleotide and the standard base
    nucleotide_matched = nucleotide[
        np.isin(nucleotide.atom_name, std_base.atom_name)
    ]
    std_base_matched = std_base[
        np.isin(std_base.atom_name, nucleotide.atom_name)
    ]
    # Ensure the nucleotide does not contain duplicate atom names
    _, unique_indices = np.unique(
        nucleotide_matched.atom_name, return_index=True
    )
    nucleotide_matched = nucleotide_matched[unique_indices]
    # Only continue if minimum number of matching atoms is reached
    if len(nucleotide_matched) < min_atoms_per_base:
        warnings.warn(
            f"Nucleotide with res_id {nucleotide.res_id[0]} and "
            f"chain_id {nucleotide.chain_id[0]} has less than 3 base "
            f"atoms, unable to check for base pair.",
            IncompleteStructureWarning
        )
        return None
    # Reorder the atoms of the nucleotide to obtain the standard RCSB
    # PDB atom order.
    nucleotide_matched = nucleotide_matched[
        standardize_order(nucleotide_matched)
    ]

    # Match the selected std_base to the base.
    _, transformation = superimpose(nucleotide_matched, std_base_matched)

    # Transform the vectors
    trans1, rot, trans2 = transformation
    vectors += trans1
    vectors  = np.dot(rot, vectors.T).T
    vectors += trans2
    # Normalize the base-normal-vector
    vectors[1,:] = vectors[1,:]-vectors[0,:]
    norm_vector(vectors[1,:])

    return vectors


def map_nucleotide(residue, min_atoms_per_base=3, rmsd_cutoff=0.28):
    """
    Map a nucleotide to one of the 5 common bases Adenine, Guanine,
    Thymine, Cytosine, and Uracil. If one of those bases bound to
    Deoxyribose and Ribose is detected as input, the corresponding one-
    letter-code (``A``, ``G``, ``T``, ``C``, ``U``) is returned.

    If a different nucleotide is given, it is mapped to the best
    fitting base using the algorithm described below.

    (i) The number of matching atom names with the reference bases is
        counted. If the number of matching atoms with all reference
        bases is less than the specified `min_atoms_per_base`
        (default 3) the nucleotide cannot be mapped and ``None`` is
        returned.

    (ii) The bases with maximum number of matching atoms are selected
         and superimposed with each reference. The base with lowest RMSD
         is chosen. If the RMSD is more than the specified
         `rmsd_cutoff` (default 0.28) the nucleotide cannot be mapped
         and ``None`` is returned.

    Parameters
    ----------
    residue : AtomArray
        The nucleotide to be mapped.
    min_atoms_per_base : int, optional (default: 3)
        The number of atoms the residue must have in common with the
        reference.
    rmsd_cutoff : float, optional (default: 0.28)
        The maximum RSMD that is allowed for a mapping to occur.

    Returns
    -------
    one_letter_code : str
        The one-letter-code of the mapped base. ``None`` if no base can
        be mapped.
    exact_match : bool
        Wether or not the residue name exactly matches one of the common
        bases, i.e. the ``res_name`` of the input `residue` is one of
        ``A``, ``G``, ``T``, ``C``, ``U``, ``DA``, ``DG``, ``DT``,
        ``DC`` or ``DU``.

    Notes
    -----
    The default RMSD cutoff was chosen according to :footcite:`Lu2015`,
    where the same cutoff is used to detect if a given base is a
    nucleotide, by superimposing the base ring atoms onto a reference
    structure.

    References
    ----------
    
    .. footbibliography::
    """
    # Check if the residue is a 'standard' nucleotide
    if residue.res_name[0] in _REFERENCE_NUCLEOTIDE_NAMES:
        return residue.res_name[0][-1], True

    # List of the standard bases for easy iteration
    std_base_list = [
        _STD_ADENINE, _STD_THYMINE, _STD_CYTOSINE, _STD_GUANINE,
        _STD_URACIL
    ]

    # The number of matched atoms for each 'standard' base
    matched_atom_no = [
        np.sum(np.isin(ref_base.atom_name, residue.atom_name))
        for ref_base in std_base_list
    ]

    if np.max(matched_atom_no) < min_atoms_per_base:
        warnings.warn(
            f"Base with res_id {residue.res_id[0]} and chain_id "
            f"{residue.chain_id[0]} has an overlap with the reference "
            f"bases which is less than {min_atoms_per_base} atoms. "
            f"Unable to map nucleotide.",
            IncompleteStructureWarning
        )
        return None, False

    # The one letter code of the best matching reference base
    best_base = None

    # Iterate through the reference bases with the maximum number of
    # matching atoms
    for ref_base in np.array(std_base_list, dtype='object')[
        np.array(matched_atom_no) == np.max(matched_atom_no)
    ]:
        # Copy the residue as the res_name property of the ``AtomArray``
        # has to be modified for later function calls.
        nuc = residue.copy()

        # Select the matching atoms of the nucleotide and the reference
        # base
        nuc = nuc[
            np.isin(nuc.atom_name, ref_base.atom_name)
        ]
        ref_base_matched = ref_base[
            np.isin(ref_base.atom_name, nuc.atom_name)
        ]

        # Set the res_name property to the same as the reference base.
        # This is a requirement for ``standardize_order``
        nuc.res_name = ref_base_matched.res_name
        # Reorder the atoms of the nucleotide to obtain the standard
        # RCSB PDB atom order. If a residue contains multiple atoms with
        # the same ``atom_name`` an exception is thrown by
        # ``standardize_order``. The exception is caught and the
        # selected reference is disregarded
        try:
            nuc = nuc[standardize_order(nuc)]
        except Exception:
            continue

        # Superimpose the nucleotide to the reference base
        fitted, _ = superimpose(ref_base_matched, nuc)

        # If the RMSD is lower than the specified cutoff or better than
        # a previous found reference, the current reference is selected
        # as best base
        if(rmsd(fitted, ref_base_matched) < rmsd_cutoff):
            rmsd_cutoff = rmsd(fitted, ref_base_matched)
            best_base = ref_base_matched.res_name[0][-1]

    if best_base is None:
        warnings.warn(
            f"Base Type {residue.res_name[0]} not supported. ",
            UnexpectedStructureWarning
        )
        return None

    return best_base, False


def _get_proximate_residues(atom_array, boolean_mask, cutoff):
    """
    Filter for residue pairs based on the distance between selected
    atoms.

    Parameters
    ----------
    atom_array : AtomArray, shape=(n,)
        The :class:`AtomArray`` to find basepair candidates in.
    boolean_mask : ndarray, dtype=bool, shape=(n,)
        The selection of atoms.
    cutoff : integer
        The maximum distance between the atoms of the two residues.

    Returns
    -------
    pairs : ndarray, dtype=int, shape=(n,2)
        Contains the basepair candidates. Each row is equivalent to one
        potential basepair. bases are represented as the first indices
        of their corresponding residues.
    count : ndarray, dtype=int, shape=(n,)
        The number of atom pairs between the residues within the
        specified cutoff
    """

    # Get the indices of the atoms that are within the maximum cutoff
    # of each other
    indices = CellList(
        atom_array, cutoff, selection=boolean_mask
    ).get_atoms(atom_array.coord[boolean_mask], cutoff)

    # Loop through the indices of potential partners
    pairs = []
    for candidate, partners in zip(np.argwhere(boolean_mask)[:, 0], indices):
        for partner in partners:
            if partner != -1:
                pairs.append((candidate, partner))

    # Get the residue starts for the indices of the candidate/partner
    # indices.
    pairs = np.array(pairs)
    basepair_candidates_shape = pairs.shape
    pairs = get_residue_starts_for(
        atom_array, pairs.flatten()
    ).reshape(basepair_candidates_shape)

    # Remove candidates where the pairs are from the same residue
    pairs = np.delete(
        pairs, np.where(
            pairs[:,0] == pairs[:,1]
        ), axis=0
    )
    # Sort the residue starts for each pair
    for i, candidate in enumerate(pairs):
        pairs[i] = sorted(candidate)
    # Make sure each pair is only listed once, count the occurrences
    pairs, count = np.unique(pairs, axis=0, return_counts=True)

    return pairs, count


def _filter_atom_type(atom_array, atom_names):
    """
    Get all atoms with specified atom names.

    Parameters
    ----------
    atom_array : AtomArray
        The :class:`AtomArray` to filter.
    atom_names : array_like
        The desired atom names.

    Returns
    -------
    filter : ndarray, dtype=bool
        This array is ``True`` for all indices in the :class:`AtomArray`
        , where the atom has the desired atom names.
    """
    return (np.isin(atom_array.atom_name, atom_names)
            & (atom_array.res_id != -1))



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import glob
import itertools
from os.path import join
import numpy as np
import pytest
import biotite.structure as struc
import biotite.structure.info as strucinfo
from biotite.structure.io import load_structure, save_structure
import biotite.structure.io.mmtf as mmtf
from ..util import data_dir


def test_mass():
    """
    Test whether the mass of a residue is the same as the sum of the
    masses of its contained atoms.
    """
    array = load_structure(join(data_dir("structure"), "1l2y.mmtf"))[0]
    _, res_names = struc.get_residues(array)
    water_mass = strucinfo.mass("H") * 2 + strucinfo.mass("O")
    # Mass of water must be subtracted
    masses = [strucinfo.mass(res_name) - water_mass for res_name in res_names]
    # C-terminus normally has additional oxygen atom
    masses[-1] += strucinfo.mass("O")
    ref_masses = [strucinfo.mass(res) for res in struc.residue_iter(array)]
    # Up to three additional/missing hydrogens are allowed
    # (protonation state)
    mass_diff = np.abs(np.array(
        [mass - ref_mass for mass, ref_mass in zip(masses, ref_masses)]
    ))
    assert (mass_diff // strucinfo.mass("H") <= 3).all()
    assert np.allclose((mass_diff % strucinfo.mass("H")), 0, atol=5e-3)


@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "*.mmtf"))
)
def test_bonds(path):
    """
    Test whether the bond data is consistent with the content of MMTF
    files.
    """
    bond_data = strucinfo.bond_dataset()
    mmtf_file = mmtf.MMTFFile.read(path)
    for group in mmtf_file["groupList"]:
        group_name = group["groupName"]
        atom_names = group["atomNameList"]
        bond_indices = group["bondAtomList"]
        bond_orders = group["bondOrderList"]
        for i in range(0, len(bond_indices), 2):
            atom1 = atom_names[bond_indices[i]]
            atom2 = atom_names[bond_indices[i+1]]
            order = bond_orders[i//2]
            ref_order = strucinfo.bond_type(group_name, atom1, atom2)
            # MMTF does not support aromaticity
            assert order == ref_order.without_aromaticity()
            assert any((
                (atom1, atom2) in strucinfo.bonds_in_residue(group_name),
                (atom2, atom1) in strucinfo.bonds_in_residue(group_name)
            ))
            assert any((
                (atom1, atom2) in bond_data[group_name],
                (atom2, atom1) in bond_data[group_name]
            ))


def test_protOr_radii():
    """
    Assert that ProtOr VdW radii (except hydrogen) can be calculated for
    all atoms in the given structure, since the structure (1GYA)
    does only contain standard amino acids after the removal of
    glycosylation.
    This means, that none of the resulting radii should be the None.
    """
    array = load_structure(join(data_dir("structure"), "1gya.mmtf"))
    array = array[..., array.element != "H"]
    array = array[..., struc.filter_amino_acids(array)]
    for res_name, atom_name in zip(array.res_name, array.atom_name):
        radius = strucinfo.vdw_radius_protor(res_name, atom_name)
        assert isinstance(radius, float)
        assert radius != None


def test_protor_radii_invalid():
    with pytest.raises(ValueError):
        # Expect raised exception for hydrogen atoms
        strucinfo.vdw_radius_protor("FOO", "H1")
    with pytest.raises(KeyError):
        # Expect raised exception when a residue does not contain an atom
        strucinfo.vdw_radius_protor("ALA", "K")
    # For all other unknown radii expect None
    assert strucinfo.vdw_radius_protor("HOH", "O") == None


def test_single_radii():
    assert strucinfo.vdw_radius_single("N") == 1.55


def test_full_name():
    assert strucinfo.full_name("Ala").upper() == "ALANINE"
    assert strucinfo.full_name("ALA").upper() == "ALANINE"


def test_link_type():
    assert strucinfo.link_type("Ala").upper() == "L-PEPTIDE LINKING"
    assert strucinfo.link_type("ALA").upper() == "L-PEPTIDE LINKING"


@pytest.mark.parametrize(
    "multi_model, seed", itertools.product([False, True], range(10))
)
def test_standardize_order(multi_model, seed):
    original = load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    if not multi_model:
        original = original[0]
    # The box is not preserved when concatenating atom arrays later
    # This would complicate the atom array equality later
    original.box = None

    # Randomly reorder the atoms in each residue
    np.random.seed(seed)
    if multi_model:
        reordered = struc.AtomArrayStack(original.stack_depth(), 0)
    else:
        reordered = struc.AtomArray(0)
    for residue in struc.residue_iter(original):
        bound = residue.array_length()
        indices = np.random.choice(
            np.arange(bound), bound,replace=False
        )
        reordered += residue[..., indices]

    # Restore the original PDB standard order
    restored = reordered[..., strucinfo.standardize_order(reordered)]

    assert restored.shape == original.shape
    assert restored[..., restored.element != "H"] \
        == original[..., original.element != "H"]



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.align"
__author__ = "Patrick Kunzmann"

import numpy as np
import numbers
import copy
import textwrap
from ..alphabet import LetterAlphabet


__all__ = ["Alignment", "get_codes", "get_symbols",
           "get_sequence_identity", "get_pairwise_sequence_identity",
           "score", "find_terminal_gaps", "remove_terminal_gaps"]


class Alignment(object):
    """
    An :class:`Alignment` object stores information about which symbols
    of *n* sequences are aligned to each other and it stores the
    corresponding alignment score.
    
    Instead of saving a list of aligned symbols, this class saves the
    original *n* sequences, that were aligned, and a so called *trace*,
    which indicate the aligned symbols of these sequences.
    The trace is a *(m x n)* :class:`ndarray` with alignment length
    *m* and sequence count *n*.
    Each element of the trace is the index in the corresponding
    sequence.
    A gap is represented by the value -1.
    
    Furthermore this class provides multiple utility functions for
    conversion into strings in order to make the alignment human
    readable.
    
    Unless an :class:`Alignment` object is the result of an multiple
    sequence alignment, the object will contain only two sequences.
    
    All attributes of this class are publicly accessible.
    
    Parameters
    ----------
    sequences : list
        A list of aligned sequences.
    trace : ndarray, dtype=int, shape=(n,m)
        The alignment trace.
    score : int, optional
        Alignment score.
    
    Attributes
    ----------
    sequences : list
        A list of aligned sequences.
    trace : ndarray, dtype=int, shape=(n,m)
        The alignment trace.
    score : int
        Alignment score.
    
    Examples
    --------
    
    >>> seq1 = NucleotideSequence("CGTCAT")
    >>> seq2 = NucleotideSequence("TCATGC")
    >>> matrix = SubstitutionMatrix.std_nucleotide_matrix()
    >>> ali = align_optimal(seq1, seq2, matrix)[0]
    >>> print(ali)
    CGTCAT--
    --TCATGC
    >>> print(ali.trace)
    [[ 0 -1]
     [ 1 -1]
     [ 2  0]
     [ 3  1]
     [ 4  2]
     [ 5  3]
     [-1  4]
     [-1  5]]
    >>> print(ali[1:4].trace)
    [[ 1 -1]
     [ 2  0]
     [ 3  1]]
    >>> print(ali[1:4, 0:1].trace)
    [[1]
     [2]
     [3]]
    """

    def __init__(self, sequences, trace, score=None):
        self.sequences = sequences.copy()
        self.trace = trace
        self.score = score

    def __repr__(self):
        """Represent Alignment a string for debugging."""
        return f"Alignment([{', '.join([seq.__repr__() for seq in self.sequences])}], " \
               f"np.{np.array_repr(self.trace)}, score={self.score})"

    def _gapped_str(self, seq_index):
        seq_str = ""
        for i in range(len(self.trace)):
            j = self.trace[i][seq_index]
            if j != -1:
                seq_str += self.sequences[seq_index][j]
            else:
                seq_str += "-"
        return seq_str
    
    def get_gapped_sequences(self):
        """
        Get a the string representation of the gapped sequences.
        
        Returns
        -------
        sequences : list of str
            The list of gapped sequence strings. The order is the same
            as in `Alignment.sequences`.
        """
        return [self._gapped_str(i) for i in range(len(self.sequences))]
    
    def __str__(self):
        # Check if any of the sequences
        # has an non-single letter alphabet
        all_single_letter = True
        for seq in self.sequences:
            if not isinstance(seq.get_alphabet(), LetterAlphabet):
                all_single_letter = False
        if all_single_letter:
            # First dimension: sequence number,
            # second dimension: line number
            seq_str_lines_list = []
            wrapper = textwrap.TextWrapper(break_on_hyphens=False)
            for i in range(len(self.sequences)):
                seq_str_lines_list.append(wrapper.wrap(self._gapped_str(i)))
            ali_str = ""
            for row_i in range(len(seq_str_lines_list[0])):
                for seq_j in range(len(seq_str_lines_list)):
                    ali_str += seq_str_lines_list[seq_j][row_i] + "\n"
                ali_str += "\n"
            # Remove final line breaks
            return ali_str[:-2]
        else:
            return super().__str__()
    
    def __getitem__(self, index):
        if isinstance(index, tuple):
            if len(index) > 2:
                raise IndexError("Only 1D or 2D indices are allowed")
            if isinstance(index[0], numbers.Integral) or \
               isinstance(index[0], numbers.Integral):
                    raise IndexError(
                        "Integers are invalid indices for alignments, "
                        "a single sequence or alignment column cannot be "
                        "selected"
                    )
            return Alignment(
                Alignment._index_sequences(self.sequences, index[1]),
                self.trace[index],
                self.score
            )
        else:
            return Alignment(self.sequences, self.trace[index], self.score)
    
    def __iter__(self):
        raise TypeError("'Alignment' object is not iterable")
    
    def __len__(self):
        return len(self.trace)
    
    def __eq__(self, item):
        if not isinstance(item, Alignment):
            return False
        if self.sequences != item.sequences:
            return False
        if not np.array_equal(self.trace, item.trace):
            return False
        if self.score != item.score:
            return False
        return True
    
    @staticmethod
    def _index_sequences(sequences, index):
        if isinstance(index, (list, tuple)) or \
            (isinstance(index, np.ndarray) and index.dtype != bool):
                return [sequences[i] for i in index]
        elif isinstance(index, np.ndarray) and index.dtype == bool:
            return [seq for seq, mask in zip(sequences, index) if mask]
        if isinstance(index, slice):
            return sequences[index]
        else:
            raise IndexError(
                f"Invalid alignment index type '{type(index).__name__}'"
            )
    
    @staticmethod
    def trace_from_strings(seq_str_list):
        """
        Create a trace from strings that represent aligned sequences.
        
        Parameters
        ----------
        seq_str_list : list of str
            The strings, where each each one represents a sequence
            (with gaps) in an alignment.
            A ``-`` is interpreted as gap.
        
        Returns
        -------
        trace : ndarray, dtype=int, shape=(n,2)
            The created trace.
        """
        if len(seq_str_list) < 2:
            raise ValueError(
                "An alignment must contain at least two sequences"
            )
        seq_i = np.zeros(len(seq_str_list))
        trace = np.full(( len(seq_str_list[0]), len(seq_str_list) ),
                        -1, dtype=int)
        # Get length of string (same length for all strings)
        # rather than length of list
        for pos_i in range(len(seq_str_list[0])):
            for str_j in range(len(seq_str_list)):
                if seq_str_list[str_j][pos_i] == "-":
                    trace[pos_i, str_j] = -1
                else:
                    trace[pos_i, str_j] = seq_i[str_j]
                    seq_i[str_j] += 1
        return trace


def get_codes(alignment):
    """
    Get the sequence codes of the sequences in the alignment.

    The codes are built from the trace:
    Instead of the indices of the aligned symbols (trace), the return
    value contains the corresponding symbol codes for each index.
    Gaps are still represented by *-1*.
    
    Parameters
    ----------
    alignment : Alignment
        The alignment to get the sequence codes for.
    
    Returns
    -------
    codes : ndarray, dtype=int, shape=(n,m)
        The sequence codes for the alignment.
        The shape is *(n,m)* for *n* sequences and *m* alignment cloumn.
        The array uses *-1* values for gaps.
    
    Examples
    --------
    
    >>> seq1 = NucleotideSequence("CGTCAT")
    >>> seq2 = NucleotideSequence("TCATGC")
    >>> matrix = SubstitutionMatrix.std_nucleotide_matrix()
    >>> ali = align_optimal(seq1, seq2, matrix)[0]
    >>> print(ali)
    CGTCAT--
    --TCATGC
    >>> print(get_codes(ali))
    [[ 1  2  3  1  0  3 -1 -1]
     [-1 -1  3  1  0  3  2  1]]
    """
    trace = alignment.trace
    sequences = alignment.sequences
    
    # The number of sequences is the first dimension
    codes = np.zeros((trace.shape[1], trace.shape[0]), dtype=int)
    for i in range(len(sequences)):
        codes[i] = np.where(
            trace[:,i] != -1, sequences[i].code[trace[:,i]], -1
        )
    
    return np.stack(codes)


def get_symbols(alignment):
    """
    Similar to :func:`get_codes()`, but contains the decoded symbols
    instead of codes.
    Gaps are still represented by *None* values.
    
    Parameters
    ----------
    alignment : Alignment
        The alignment to get the symbols for.
    
    Returns
    -------
    symbols : list of list
        The nested list of symbols.
    
    See Also
    --------
    get_codes

    Examples
    --------
    
    >>> seq1 = NucleotideSequence("CGTCAT")
    >>> seq2 = NucleotideSequence("TCATGC")
    >>> matrix = SubstitutionMatrix.std_nucleotide_matrix()
    >>> ali = align_optimal(seq1, seq2, matrix)[0]
    >>> print(ali)
    CGTCAT--
    --TCATGC
    >>> print(get_symbols(ali))
    [['C', 'G', 'T', 'C', 'A', 'T', None, None], [None, None, 'T', 'C', 'A', 'T', 'G', 'C']]
    """
    codes = get_codes(alignment)
    symbols = [None] * codes.shape[0]
    for i in range(codes.shape[0]):
        alphabet = alignment.sequences[i].get_alphabet()
        codes_wo_gaps = codes[i, codes[i] != -1]
        symbols_wo_gaps = alphabet.decode_multiple(codes_wo_gaps)
        if not isinstance(symbols_wo_gaps, list):
            symbols_wo_gaps = list(symbols_wo_gaps)
        symbols_for_seq = np.full(len(codes[i]), None, dtype=object)
        symbols_for_seq[codes[i] != -1] = symbols_wo_gaps
        symbols[i] = symbols_for_seq.tolist()
    return symbols


def get_sequence_identity(alignment, mode="not_terminal"):
    """
    Calculate the sequence identity for an alignment.

    The identity is equal to the matches divided by a measure for the
    length of the alignment that depends on the `mode` parameter.
    
    Parameters
    ----------
    alignment : Alignment
        The alignment to calculate the identity for.
    mode : {'all', 'not_terminal', 'shortest'}, optional
        The calculation mode for alignment length.

            - **all** - The number of matches divided by the number of
              all alignment columns.
            - **not_terminal** - The number of matches divided by the
              number of alignment columns that are not terminal gaps in
              any of the sequences.
            - **shortest** - The number of matches divided by the
              length of the shortest sequence.

        Default is *not_terminal*.
    
    Returns
    -------
    identity : float
        The sequence identity, ranging between 0 and 1.
    
    See also
    --------
    get_pairwise_sequence_identity
    """
    codes = get_codes(alignment)

    # Count matches
    matches = 0
    for i in range(codes.shape[1]):
        column = codes[:,i]
        # One unique value -> all symbols match
        unique_symbols = np.unique(column)
        if len(unique_symbols) == 1 and unique_symbols[0] != -1:
            matches += 1
    
    # Calculate length
    if mode == "all":
        length = len(alignment)
    elif mode == "not_terminal":
        start, stop = find_terminal_gaps(alignment)
        if stop <= start:
            raise ValueError(
                "Cannot calculate non-terminal identity, "
                "at least two sequences have no overlap"
            )
        length = stop - start
    elif mode == "shortest":
        length = min([len(seq) for seq in alignment.sequences])
    else:
        raise ValueError(f"'{mode}' is an invalid calculation mode")

    return matches / length


def get_pairwise_sequence_identity(alignment, mode="not_terminal"):
    """
    Calculate the pairwise sequence identity for an alignment.

    The identity is equal to the matches divided by a measure for the
    length of the alignment that depends on the `mode` parameter.
    
    Parameters
    ----------
    alignment : Alignment, length=n
        The alignment to calculate the pairwise sequence identity for.
    mode : {'all', 'not_terminal', 'shortest'}, optional
        The calculation mode for alignment length.

            - **all** - The number of matches divided by the number of
              all alignment columns.
            - **not_terminal** - The number of matches divided by the
              number of alignment columns that are not terminal gaps in
              any of the two considered sequences.
            - **shortest** - The number of matches divided by the
              length of the shortest one of the two sequences.

        Default is *not_terminal*.
    
    Returns
    -------
    identity : ndarray, dtype=float, shape=(n,n)
        The pairwise sequence identity, ranging between 0 and 1.
    
    See also
    --------
    get_sequence_identity
    """
    codes = get_codes(alignment)
    n_seq = len(codes)

    # Count matches
    # Calculate at which positions the sequences are identical
    # and are not gaps
    equality_matrix = (codes[:, np.newaxis, :] == codes[np.newaxis, :, :]) \
                    & (codes[:, np.newaxis, :] != -1) \
                    & (codes[np.newaxis, :, :] != -1) \
    # Sum these positions up
    matches = np.count_nonzero(equality_matrix, axis=-1)

    # Calculate length
    if mode == "all":
        length = len(alignment)
    elif mode == "not_terminal":
        length = np.zeros((n_seq, n_seq))
        for i in range(n_seq):
            for j in range(n_seq):
                # Find latest start and earliest stop of all sequences
                start, stop = find_terminal_gaps(alignment[:, [i,j]])
                if stop <= start:
                    raise ValueError(
                        "Cannot calculate non-terminal identity, "
                        "as the two sequences have no overlap"
                    )
                length[i,j] = stop - start
    elif mode == "shortest":
        length = np.zeros((n_seq, n_seq))
        for i in range(n_seq):
            for j in range(n_seq):
                length[i,j] = min([
                    len(alignment.sequences[i]),
                    len(alignment.sequences[j])
                ])
    else:
        raise ValueError(f"'{mode}' is an invalid calculation mode")
    
    return matches / length


def score(alignment, matrix, gap_penalty=-10, terminal_penalty=True):
    """
    Calculate the similarity score of an alignment.

    If the alignment contains more than two sequences,
    all pairwise scores are counted.
    
    Parameters
    ----------
    alignment : Alignment
        The alignment to calculate the identity for.
    matrix : SubstitutionMatrix
        The substitution matrix used for scoring.
    gap_penalty : int or (tuple, dtype=int), optional
        If an integer is provided, the value will be interpreted as
        general gap penalty. If a tuple is provided, an affine gap
        penalty is used. The first integer in the tuple is the gap
        opening penalty, the second integer is the gap extension
        penalty.
        The values need to be negative. (Default: *-10*)
    terminal_penalty : bool, optional
        If true, gap penalties are applied to terminal gaps.
        (Default: True)
    
    Returns
    -------
    score : int
        The similarity score.
    """
    codes = get_codes(alignment)
    matrix = matrix.score_matrix()

    # Sum similarity scores (without gaps)
    score = 0
    # Iterate over all positions
    for pos in range(codes.shape[1]):
        column = codes[:, pos]
        # Iterate over all possible pairs
        # Do not count self-similarity
        # and do not count similarity twice (not S(i,j) and S(j,i))
        for i in range(codes.shape[0]):
            for j in range(i+1, codes.shape[0]):
                code_i = column[i]
                code_j = column[j]
                # Ignore gaps
                if code_i != -1 and code_j != -1:
                    score += matrix[code_i, code_j]
    
    # Sum gap penalties
    if type(gap_penalty) == int:
        gap_open = gap_penalty
        gap_ext = gap_penalty
    elif type(gap_penalty) == tuple:
        gap_open = gap_penalty[0]
        gap_ext = gap_penalty[1]
    else:
        raise TypeError("Gap penalty must be either integer or tuple")
    # Iterate over all sequences
    for seq_code in codes:
        in_gap = False
        if terminal_penalty:
            start_index = 0
            stop_index = len(seq_code)
        else:
            # Find a start and stop index excluding terminal gaps
            start_index, stop_index = find_terminal_gaps(alignment)
        for i in range(start_index, stop_index):
            if seq_code[i] == -1:
                if in_gap:
                    score += gap_ext
                else:
                    score += gap_open
                in_gap = True
            else:
                in_gap = False
    return score


def find_terminal_gaps(alignment):
    """
    Find the slice indices that would remove terminal gaps from an
    alignment.

    Terminal gaps are gaps that appear before all sequences start and
    after any sequence ends.

    Parameters
    ----------
    alignment : Alignment
        The alignment, where the slice indices should be found in.

    Returns
    -------
    start, stop : int
        Indices that point to the start and exclusive stop of the
        alignment columns without terminal gaps.
        When these indices are used as slice index for an alignment or
        trace, the index would remove terminal gaps.

    See also
    --------
    remove_terminal_gaps

    Examples
    --------

    >>> sequences = [
    ...     NucleotideSequence(seq_string) for seq_string in (
    ...         "AAAAACTGATTC",
    ...         "AAACTGTTCA",
    ...         "CTGATTCAAA"
    ...     )
    ... ]
    >>> trace = np.transpose([
    ...     ( 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, -1, -1, -1),
    ...     (-1, -1,  0,  1,  2,  3,  4,  5, -1,  6,  7,  8,  9, -1, -1),
    ...     (-1, -1, -1, -1, -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9),
    ... ])
    >>> alignment = Alignment(sequences, trace)
    >>> print(alignment)
    AAAAACTGATTC---
    --AAACTG-TTCA--
    -----CTGATTCAAA
    >>> print(find_terminal_gaps(alignment))
    (5, 12)
    """
    trace = alignment.trace
    # Find for each sequence the positions of non-gap symbols
    no_gap_pos = [np.where(trace[:,i] != -1)[0] for i in range(trace.shape[1])]
    # Find for each sequence the positions of the sequence start and end
    # in the alignment
    firsts = [no_gap_pos[i][0 ] for i in range(trace.shape[1])]
    lasts  = [no_gap_pos[i][-1] for i in range(trace.shape[1])]
    # The terminal gaps are before all sequences start and after any
    # sequence ends
    # Use exclusive stop -> -1
    return np.max(firsts), np.min(lasts) + 1


def remove_terminal_gaps(alignment):
    """
    Remove terminal gaps from an alignment.

    Terminal gaps are gaps that appear before all sequences start and
    after any sequence ends.

    Parameters
    ----------
    alignment : Alignment
        The alignment, where the terminal gaps should be removed from.

    Returns
    -------
    truncated_alignment : Alignment
        A shallow copy of the input `alignment` with an truncated trace,
        that does not contain alignment columns with terminal gaps.

    See also
    --------
    find_terminal_gaps

    Examples
    --------

    >>> sequences = [
    ...     NucleotideSequence(seq_string) for seq_string in (
    ...         "AAAAACTGATTC",
    ...         "AAACTGTTCA",
    ...         "CTGATTCAAA"
    ...     )
    ... ]
    >>> trace = np.transpose([
    ...     ( 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, -1, -1, -1),
    ...     (-1, -1,  0,  1,  2,  3,  4,  5, -1,  6,  7,  8,  9, -1, -1),
    ...     (-1, -1, -1, -1, -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9),
    ... ])
    >>> alignment = Alignment(sequences, trace)
    >>> print(alignment)
    AAAAACTGATTC---
    --AAACTG-TTCA--
    -----CTGATTCAAA
    >>> truncated_alignment = remove_terminal_gaps(alignment)
    >>> print(truncated_alignment)
    CTGATTC
    CTG-TTC
    CTGATTC
    """
    start, stop = find_terminal_gaps(alignment)
    if stop < start:
        raise ValueError(
            "Cannot remove terminal gaps, since at least two sequences have "
            "no overlap and the resulting alignment would be empty"
        )
    return alignment[start : stop]


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
Some convenience functions for consistency with other ``structure.io``
subpackages.
"""

__name__ = "biotite.structure.io.pdbqt"
__author__ = "Patrick Kunzmann"
__all__ = ["get_structure", "set_structure"]


def get_structure(pdbqt_file, model=None):
    """
    Get an :class:`AtomArray` or :class:`AtomArrayStack` from the
    PDBQT file.

    EXPERIMENTAL: Future API changes are probable.
    
    Parameters
    ----------
    pdbqt_file : PDBQTFile
        The PDBQT file.
    model : int, optional
        If this parameter is given, the function will return an
        :class:`AtomArray` from the atoms corresponding to the given
        model number (starting at 1).
        Negative values are used to index models starting from the
        last model insted of the first model.
        If this parameter is omitted, an :class:`AtomArrayStack`
        containing all models will be returned, even if the
        structure contains only one model.
    
    Returns
    -------
    array : AtomArray or AtomArrayStack
        The return type depends on the `model` parameter.
    """
    return pdbqt_file.get_structure(model)


def set_structure(pdbqt_file, atoms, charges=None, atom_types=None,
                  rotatable_bonds=None, root=None, include_torsdof=True):
    """
    Write an :class:`AtomArray` into a PDBQT file.

    EXPERIMENTAL: Future API changes are probable.
    
    Parameters
    ----------
    pdbqt_file : PDBQTFile
        The PDBQT file.
    atoms : AtomArray, shape=(n,)
        The atoms to be written into this file.
        Must have an associated :class:`BondList`.
    charges : ndarray, shape=(n,), dtype=float, optional
        Partial charges for each atom in `atoms`.
        By default, the charges are calculated using the PEOE method
        (:func:`partial_charges()`).
    atom_types : ndarray, shape=(n,), dtype="U1", optional
        Custom *AutoDock* atom types for each atom in `atoms`.
    rotatable_bonds : None or 'rigid' or 'all' or BondList, optional
        This parameter describes, how rotatable bonds are handled,
        with respect to ``ROOT``, ``BRANCH`` and ``ENDBRANCH``
        lines.

            - ``None`` - The molecule is handled as rigid receptor:
              No ``ROOT``, ``BRANCH`` and ``ENDBRANCH`` lines will
              be written.
            - ``'rigid'`` - The molecule is handled as rigid ligand:
              Only a ``ROOT`` line will be written.
            - ``'all'`` - The molecule is handled as flexible 
              ligand:
              A ``ROOT`` line will be written and all rotatable
              bonds are included using ``BRANCH`` and ``ENDBRANCH``
              lines.
            - :class:`BondList` - The molecule is handled as
              flexible ligand:
              A ``ROOT`` line will be written and all bonds in the
              given :class:`BondList` are considered flexible via
              ``BRANCH`` and ``ENDBRANCH`` lines.
        
    root : int, optional
        Specifies the index of the atom following the ``ROOT`` line.
        Setting the root atom is useful for specifying the *anchor*
        in flexible side chains.
        This parameter has no effect, if `rotatable_bonds` is
        ``None``.
        By default, the first atom is also the root atom.
    include_torsdof : bool, optional
        By default, a ``TORSDOF`` (torsional degrees of freedom)
        record is written at the end of the file.
        By setting this parameter to false, the record is omitted.
    
    Returns
    -------
    mask : ndarray, shape=(n,), dtype=bool
        A boolean mask, that is ``False`` for each atom of the input
        ``atoms``, that was removed due to being a nonpolar
        hydrogen.
    """
    return pdbqt_file.set_structure(
        atoms, charges, atom_types, rotatable_bonds, root,
        include_torsdof
    )


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for visualization of sequence related objects.

The visualizations make use of *Matplotlib* for plotting.
Therefore, each plotting function requires an :class:`Axes` object,
where the visualization should be performed.

The resulting plots are customizable:
Labels, location numbers, etc. are usually placed into the axes tick
labels, making tham accessible for modification via the usual
*Matplotlib* API.
Some aspects of the plotting itself are also customizable: For example
the :func:`plot_alignment()` function uses a interchangable
:class:`SymbolPlotter`, that knows how to draw the symbols of an
alignment.
Similarily, the appearance of sequence features in the function
:func:`plot_feature_map()` is customized via :class:`FeaturePlotter`
objects.
"""

__name__ = "biotite.sequence.graphics"
__author__ = "Patrick Kunzmann"

from .alignment import *
from .colorschemes import *
from .dendrogram import *
from .features import *
from .plasmid import *
from .logo import *



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.info"
__author__ = "Tom David Müller"
__all__ = ["nucleotide_names"]

import json
import numpy as np
from os.path import join, dirname, realpath


_info_dir = dirname(realpath(__file__))
# Data is taken from
# ftp://ftp.wwpdb.org/pub/pdb/data/monomers/components.cif
# (2020/10/21)
# The json-file contains all three-letter-codes of the components where
# the data item `_chem_comp.type` is equal to one of the following
# values:
# DNA OH 3 prime terminus, DNA OH 5 prime terminus, DNA linking,
# RNA OH 3 prime terminus, RNA OH 5 prime terminus, RNA linking,
# L-RNA LINKING, L-DNA LINKING
with open(join(_info_dir, "nucleotides.json"), "r") as file:
    _nucleotides = json.load(file)

def nucleotide_names():
    """
    Get a list of nucleotide three-letter codes according to the PDB
    chemical compound dictionary.

    Returns
    -------
    nucleotide_names : list
        A list of three-letter-codes containing residues that are
        DNA/RNA-Linking.
    """
    return _nucleotides



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from os.path import join
import pytest
import biotite.sequence as seq
import biotite.sequence.io.fasta as fasta
from ...util import data_dir


@pytest.fixture
def sequences():
    """
    10 Cas9 sequences.
    """
    fasta_file = fasta.FastaFile.read(join(data_dir("sequence"), "cas9.fasta"))
    return [seq.ProteinSequence(sequence) for sequence in fasta_file.values()]